# -*- coding: utf-8 -*-
"""
Created on Feb 2018

@author: Chester (Yu-Chuan Chang)
"""

""""""""""""""""""""""""""""""
# import libraries
""""""""""""""""""""""""""""""
import os
import sys
import numpy as np

""""""""""""""""""""""""""""""
# define functions 
""""""""""""""""""""""""""""""
def EstimateAlleleFrequency(gen_snp):
    """

    A function for estimating allele frequency of a single varaint

    Args:
        gen_snp (list): The genotypes of a variant of all samples
        
    Returns:
        (tuple): tuple containing:

            - float_frequency_A (float): The reference allele type frequency
            - float_frequency_B (float): The alternative allele type frequency
    
    """
    
    ### get all subject's genotype
    list_snp = gen_snp.split(" ")[5:]
    ### get the number of subjects
    int_num_subject = int(len(list_snp)/3)
    
    ### generate count table (AA, AB, BB)
    list_count = [0, 0, 0]
    for idx_subject in range(0, int_num_subject):
        idx_col = np.argmax(list_snp[idx_subject * 3: idx_subject * 3 + 3])
        list_count[idx_col] = list_count[idx_col] + 1
    
    ### calculate allele frequency
    ### frequency of A = AA + AB/2
    ### frequency of B = BB + AB/2
    float_frequency_A = float(list_count[0] + list_count[1] / 2) / int_num_subject
    float_frequency_B = float(list_count[2] + list_count[1] / 2) / int_num_subject
    
    return float_frequency_A, float_frequency_B

def EstimatePairwiseLD(gen_snp_1, gen_snp_2):
    """

    Lewontin (1964) linkage disequilibrium (LD) estimation.

    Args:
        gen_snp_1 (list): The genotypes of first variant of all samples
        gen_snp_2 (list): The genotypes of second variant of all samples

    Returns:
        (tuple): tuple containing:

            - float_D_prime (float): The DPrime of these two variants
            - float_R_square (float): The RSquare of these two variants
    
    """

    ### get all subject's genotype
    list_snp1 = gen_snp_1.split(" ")[5:]
    list_snp2 = gen_snp_2.split(" ")[5:]
    ### get the number of subjects
    int_num_subject = int(len(list_snp1)/3)
    
    ### generate contigency table
    ### row: SNP1_AA; SNP1_Aa; SNP1_aa
    ### col: SNP2_bb; SNP2_Bb; SNP2_bb
    np_contigency = np.zeros((3, 3), np.dtype(int))
    for idx_subject in range(0, int_num_subject):
        idx_row = np.argmax(np.array(list_snp1[idx_subject * 3: idx_subject * 3 + 3]))
        idx_col = np.argmax(np.array(list_snp2[idx_subject * 3: idx_subject * 3 + 3]))
        np_contigency[idx_row, idx_col] = np_contigency[idx_row, idx_col] + 1
    
    ### estimate single locus haplotyes
    ### snp1_A = (AABB + AABb + AAbb) + (AaBB + AaBb + Aabb)/2; snp1_a = snp1_A - 1
    float_probability_A = float(np.sum(np_contigency[0, :]) + float(np.sum(np_contigency[1, :])) / 2) / int_num_subject
    float_probability_a = 1 - float_probability_A
    ### snp2_B = (AABB + AaBB + aaBB) + (AABb + AaBb + aaBb)/2; snp2_b = snp2_B - 1
    float_probability_B = float(np.sum(np_contigency[:, 0]) + float(np.sum(np_contigency[:, 1])) / 2) / int_num_subject
    float_probability_b = 1 - float_probability_B
    
    ### set arbitrary probability of AB
    float_probability_AB = float_probability_A * float_probability_B
    
    try:
        ### EM algorithm
        for idx_loop in range(0, 10000):
            ### E(num_AB|prob_AB) = 2 * num_AABB + num_AABb + num_AaBB +
            ### (prob_AB * (1 + prob_AB - prob_A - prob_B) * num_AbBb) / 
            ### ((prob_A - prob_AB) * (prob_B - prob_AB) + prob_AB * (1 + prob_AB - prob_A - prob_B))
            float_num_AB_estimateByEM = 2 * float(np_contigency[0, 0]) + float(np_contigency[0, 1]) + float(np_contigency[1, 0]) + (float_probability_AB * (1 + float_probability_AB - float_probability_A - float_probability_B) * float(np_contigency[1, 1])) / ((float_probability_A - float_probability_AB) * (float_probability_B - float_probability_AB) + float_probability_AB * (1 + float_probability_AB - float_probability_A - float_probability_B))
            float_probability_AB_estimateByEM = float_num_AB_estimateByEM / (int_num_subject * 2)
            if abs(float_probability_AB_estimateByEM - float_probability_AB) < 0.0000001:
                break
            else:
                float_probability_AB = float_probability_AB_estimateByEM
        
        ### calculate D
        float_D = float_probability_AB - float_probability_A * float_probability_B
        ### calculate D prime
        if float_D >= 0:
            float_D_min = min([float_probability_A * (1 - float_probability_B), (1 - float_probability_A) * float_probability_B])
        else:
            float_D_min = max([-float_probability_A * float_probability_B, -(1 - float_probability_A) * (1 - float_probability_B)])
        float_D_prime = float_D / float_D_min
        ### calculate R square
        float_R_square = (float_D**2) / (float_probability_A * float_probability_a * float_probability_B * float_probability_b)
        
        return float_D_prime, float_R_square
    
    except ZeroDivisionError:
        return 1.0, 1.0

""""""""""""""""""""""""""""""
# main function
""""""""""""""""""""""""""""""
def EstimateLDBlock(str_inputFileName_genotype, str_outputFilePath = "", float_threshold_DPrime = 0.8, float_threshold_RSquare = 0.8):
    """

    A function for implementing linkage disequilibrium (LD) dimension reduction. In genotype data, a variant often exhibits high dependency with its nearby variants because of LD. In the practical implantation, we prefer to group these dependent features to reduce the dimension of features. In other words, we can take the advantages of LD to reduce the dimensionality of genetic features. In this regard, this function adopted the same approach developed by Lewontin (1964) to estimate LD. We used D’ and r2 as the criteria to group highly dependent genetic features as blocks. In each block, we chose the features with the largest minor allele frequency to represent other features in the same block.

    Args:
        str_inputFileName_genotype (str): File name of input genotype data
        str_outputFilePath (str): File path of output file
        float_threshold_DPrime (float): The Dprime threshold for discriminating a LD block (default: 0.8)
        float_threshold_RSquare (float): The RSquare threshold for discriminating a LD block (default: 0.8)

    Returns:
        - Expected Success Response::

            "step2: Estimate LD. DONE!"
    
    """
    
    ### set default output path
    if str_outputFilePath == "":
        str_outputFilePath = os.path.dirname(str_inputFileName_genotype)
    
    ### get the number of snp
    int_num_snp = sum(1 for line in open(str_inputFileName_genotype))
    
    ### read .gen file and estimate the LD block
    list_outputLDBlock = []
    with open(str_inputFileName_genotype, "r") as file_inputFile:
        with open(os.path.join(str_outputFilePath, os.path.basename(str_inputFileName_genotype).replace(".gen", "_LDReduced.gen")), "w") as file_outputFile:
            ### create dictionary for LD block
            ### key: rsID; value:[minor allele requency, raw genotypes data]
            dict_thisLDBlock = {}
            ### put first snp into dictionary
            line_previousSnp = file_inputFile.readline()
            list_previousSnp = line_previousSnp.strip().split(" ")
            dict_thisLDBlock[list_previousSnp[1]] = [min(EstimateAlleleFrequency(line_previousSnp)), line_previousSnp]
            
            ### scan all other snps
            int_count_snp = 1
            for line in file_inputFile:
                list_thisSnp = line.strip().split(" ")
                
                ### estimate pairwise LD for all of the snps in dictionary
                bool_flag_inLD = True
                for key in dict_thisLDBlock.keys():
                    float_DPrime, float_RSquare = EstimatePairwiseLD(dict_thisLDBlock[key][1], line)
                    if float_DPrime < float_threshold_DPrime or float_RSquare < float_threshold_RSquare:
                        bool_flag_inLD = False
                        break
                
                ### if this snp not in this LD block, then output and clear the content of dictionary
                if bool_flag_inLD == False:
                    ### find a snp with maximum minor allele frequency to be representative snp
                    str_representative_rsid = list(dict_thisLDBlock.keys())[0]
                    for key in dict_thisLDBlock.keys():
                        if dict_thisLDBlock[key][0] > dict_thisLDBlock[str_representative_rsid][0]:
                            str_representative_rsid = key
                    list_outputLDBlock.append(str_representative_rsid + ":" + ",".join(dict_thisLDBlock.keys()))
                    file_outputFile.writelines(dict_thisLDBlock[str_representative_rsid][1])
                    dict_thisLDBlock.clear()
                ### add this snp to current dictionary
                dict_thisLDBlock[list_thisSnp[1]] = [min(EstimateAlleleFrequency(line)), line]
                
                ### show progress
                int_count_snp = int_count_snp + 1
                str_print = "step2: Processing: " + "{0:.2f}".format(float(int_count_snp) / int_num_snp * 100) + "%"
                sys.stdout.write('%s\r' % str_print)
                sys.stdout.flush()
            
            ### output the final LD block in dictionary
            ### find a snp with maximum minor allele frequency to be representative snp
            str_representative_rsid = list(dict_thisLDBlock.keys())[0]
            for key in dict_thisLDBlock.keys():
                if dict_thisLDBlock[key][0] > dict_thisLDBlock[str_representative_rsid][0]:
                    str_representative_rsid = key
            list_outputLDBlock.append(str_representative_rsid + ":" + ",".join(dict_thisLDBlock.keys()))
            file_outputFile.writelines(dict_thisLDBlock[str_representative_rsid][1])
    
    ### output the file of LD block
    ### output file format: rsid_representative: rsid_1,rsid_2,rsid_3,...(the snps in the same LD block)
    with open(os.path.join(str_outputFilePath, os.path.basename(str_inputFileName_genotype).replace(".gen", ".LDBlock")), "w") as file_outputFile:
        for item in list_outputLDBlock:
            file_outputFile.writelines(item + "\n")
    
    print("step2: Estimate LD. DONE! \t\t\t\t")


# -*- coding: utf-8 -*-
"""
Created on Feb 2018

@author: Chester (Yu-Chuan Chang)
"""

""""""""""""""""""""""""""""""
# import libraries
""""""""""""""""""""""""""""""
import os
import warnings
warnings.filterwarnings('ignore')
# ignore all warnings
warnings.simplefilter("ignore")
os.environ["PYTHONWARNINGS"] = "ignore"

import sys
import itertools
import numpy as np
np.seterr(divide='ignore', invalid='ignore')
from sklearn.feature_selection import VarianceThreshold
from sklearn.feature_selection import f_regression
from scipy.sparse import coo_matrix
from sklearn.utils import shuffle
from sklearn import linear_model
from sklearn.model_selection import KFold
from sklearn.model_selection import GridSearchCV
import scipy.stats as stats
import multiprocessing as mp

from genepi.tools import randomized_l1

""""""""""""""""""""""""""""""
# define functions 
""""""""""""""""""""""""""""""
def RandomizedLassoRegression(np_X, np_y):
    """

    Implementation of the stability selection.

    Args:
        np_X (ndarray): 2D array containing genotype data with `int8` type
        np_y (ndarray): 2D array containing phenotype data with `float` type

    Returns:
        (ndarray): estimator.scores 
        
            1D array containing the scores of each genetic features with `float` type
    
    """

    X = np_X
    y = np_y
    X_sparse = coo_matrix(X)
    X, X_sparse, y = shuffle(X, X_sparse, y, random_state=0)
    estimator = randomized_l1.RandomizedLasso(n_jobs=1, n_resampling=100)
    estimator.fit(X, y)
    
    return estimator.scores_

def LassoRegressionCV(np_X, np_y, int_kOfKFold = 2, int_nJobs = 1):
    """

    Implementation of the L1-regularized Lasso regression with k-fold cross validation.

    Args:
        np_X (ndarray): 2D array containing genotype data with `int8` type
        np_y (ndarray): 2D array containing phenotype data with `float` type
        int_kOfKFold (int): The k for k-fold cross validation (default: 2)
        int_nJobs (int): The number of thread (default: 1)

    Returns:
        (ndarray): estimator.scores 
        
            1D array containing the scores of each genetic features with `float` type
    
    """

    X = np_X
    y = np_y
    X_sparse = coo_matrix(X)
    X, X_sparse, y = shuffle(X, X_sparse, y, random_state=0)
    kf = KFold(n_splits=int_kOfKFold)
    
    list_target = []
    list_predict = []
    list_weight = []
    for idxTr, idxTe in kf.split(X):
        alpha = np.logspace(-10, 10, 200)
        parameters = [{'alpha':alpha}]
        kf_estimator = KFold(n_splits=2)
        estimator_lasso = linear_model.Lasso(max_iter=1000)
        estimator_grid = GridSearchCV(estimator_lasso, parameters, scoring='neg_mean_squared_error', n_jobs=1, cv=kf_estimator)
        estimator_grid.fit(X[idxTr], y[idxTr])
        list_label = estimator_grid.best_estimator_.predict(X[idxTe])
        list_weight.append([float(item) for item in estimator_grid.best_estimator_.coef_])
        for idx_y, idx_label in zip(list(y[idxTe]), list_label):
            list_target.append(float(idx_y))
            list_predict.append(idx_label)
    np_weight = np.array(list_weight)
    np_weight = np.average(list_weight, axis=0)
    float_pearson = stats.stats.pearsonr(list_target, list_predict)[0]
    float_spearman = stats.stats.spearmanr(list_target, list_predict)[0]
    
    return (float_pearson + float_spearman) / 2, np_weight

def FeatureEncoderLasso(np_genotype_rsid, np_genotype, np_phenotype, int_dim):
    """

    Implementation of the two-element combinatorial encoding.

    Args:
        np_genotype_rsid (ndarray): 1D array containing rsid of genotype data with `str` type
        np_genotype (ndarray): 2D array containing genotype data with `int8` type
        np_phenotype (ndarray): 2D array containing phenotype data with `float` type
        int_dim (int): The dimension of a variant (default: 3. AA, AB and BB)

    Returns:
        (tuple): tuple containing:

            - list_interaction_rsid (ndarray): 1D array containing rsid of genotype data with `str` type
            - np_interaction (ndarray): 2D array containing genotype data with `int8` type
    
    """

    ### combinatorial encoding
    np_interaction = np_genotype
    list_interaction_rsid = list(np_genotype_rsid)    

    list_combs = list(itertools.combinations(range(int(np_interaction.shape[1]/int_dim)), 2))
    
    for idx_combs in range(len(list_combs)):
        try:
            ### generate interaction terms
            tuple_comb = list_combs[idx_combs]
            np_this_interaction = np.zeros([np_phenotype.shape[0], int_dim**2], dtype='int8')
            list_this_interaction_id = []
            for idx_x in range(int_dim):
                for idx_y in range(int_dim):
                    np_this_interaction_term = (np_genotype[:, tuple_comb[0] * int_dim + idx_x] * np_genotype[:, tuple_comb[1] * int_dim + idx_y]).astype(np.int8)
                    if not(np.array_equal(np_this_interaction_term, np_genotype[:, tuple_comb[0] * int_dim + idx_x])) and not(np.array_equal(np_this_interaction_term, np_genotype[:, tuple_comb[1] * int_dim + idx_y])):
                        np_this_interaction[:, idx_x * int_dim + idx_y] = np_this_interaction_term
                    list_this_interaction_id.append(np_genotype_rsid[tuple_comb[0] * int_dim + idx_x] + "*" + np_genotype_rsid[tuple_comb[1] * int_dim + idx_y])
            
            ### variance check (detect variance < 0.05)
            sk_variance = VarianceThreshold(threshold=(.95 * (1 - .95)))
            np_this_interaction = sk_variance.fit_transform(np_this_interaction)
            np_this_interaction_id = np.array(list_this_interaction_id)
            np_this_interaction_id = np.array(np_this_interaction_id[sk_variance.get_support()])
            
            ### f regression feature selection
            np_fRegression = -np.log10(f_regression(np_this_interaction.astype(int), np_phenotype[:, -1].astype(float))[1])
            np_selectedIdx = np.array([x > 2 for x in np_fRegression])
            np_this_interaction = np_this_interaction[:, np_selectedIdx]
            np_this_interaction_id = np_this_interaction_id[np_selectedIdx]
        
            ### append insteraction terms
            int_num_interaction = np_this_interaction.shape[1]
            if int_num_interaction == 0:
                continue
            np_interaction_append = np.empty((np_interaction.shape[0], np_interaction.shape[1] + int_num_interaction), dtype='int')
            np_interaction_append[:,:-(int_num_interaction)] = np_interaction
            np_interaction_append[:,-(int_num_interaction):] = np_this_interaction
            np_interaction = np_interaction_append
            list_interaction_rsid.extend(list(np_this_interaction_id))
        except:
            pass

    return np.array(list_interaction_rsid), np_interaction

def FilterInLoading(np_genotype, np_phenotype):
    """

    This function is for filtering low quality varaint. Before modeling each subset of genotype features, two criteria were adopted to exclude low quality data. The first criterion is that the genotype frequency of a feature should exceed 5%, where the genotype frequency means the proportion of genotype among the total samples in the dataset. The second criterion is regarding the association between the feature and the phenotype. We used χ2 test to estimate the association between the feature and the phenotype, and the p-value should be smaller than 0.01.

    Args:
        np_genotype (ndarray): 2D array containing genotype data with `int8` type
        np_phenotype (ndarray): 2D array containing phenotype data with `float` type

    Returns:
        (ndarray): np_genotype
        
            2D array containing genotype data with `int8` type
    
    """

    try:
        ### variance check (detect variance < 0.05)
        sk_variance = VarianceThreshold(threshold=(.95 * (1 - .95)))
        np_genotype = sk_variance.fit_transform(np_genotype)
        
        ### f regression feature selection
        np_fRegression = -np.log10(f_regression(np_genotype.astype(int), np_phenotype[:, -1].astype(float))[1])
        np_selectedIdx = np.array([x > 2 for x in np_fRegression])
        np_genotype = np_genotype[:, np_selectedIdx]
        
        return np_genotype.shape[1]
    
    except:
        return 0

""""""""""""""""""""""""""""""
# main function
""""""""""""""""""""""""""""""
def SingleGeneEpistasisLasso(str_inputFileName_genotype, str_inputFileName_phenotype, str_outputFilePath = "", int_kOfKFold = 2, int_nJobs = 1):    
    """

    A workflow to model a single gene containing two-element combinatorial encoding, stability selection, filtering low quality varaint and  L1-regularized Lasso regression with k-fold cross validation.

    Args:
        str_inputFileName_genotype (str): File name of input genotype data
        str_inputFileName_phenotype (str): File name of input phenotype data
        str_outputFilePath (str): File path of output file
        int_kOfKFold (int): The k for k-fold cross validation (default: 2)
        int_nJobs (int): The number of thread (default: 1)

    Returns:
        (float): float_AVG_S_P
        
            The average of the Peason's and Spearman's correlation of the model
    
    """
    
    ### set path of output file
    if str_outputFilePath == "":
        str_outputFilePath = os.path.dirname(str_inputFileName_genotype)
    
    #-------------------------
    # load data
    #-------------------------
    ### count lines of input files
    int_num_phenotype = sum(1 for line in open(str_inputFileName_phenotype))
    
    ### get phenotype file
    list_phenotype = []
    with open(str_inputFileName_phenotype, 'r') as file_inputFile:
        for line in file_inputFile:
            list_phenotype.append(line.strip().split(","))
    np_phenotype = np.array(list_phenotype, dtype=np.float)
    del list_phenotype
    
    ### get genotype file
    list_genotype = [[] for x in range(int_num_phenotype)]
    list_genotype_rsid = []
    with open(str_inputFileName_genotype, 'r') as file_inputFile:
        for line in file_inputFile:
            list_thisSnp = line.strip().split(" ")
            np_this_genotype = np.empty([int_num_phenotype, 3], dtype='int8')
            for idx_subject in range(0, int_num_phenotype):
                list_allelType = [0, 0, 0]
                list_allelType[np.argmax(list_thisSnp[idx_subject * 3 + 5 : idx_subject * 3 + 8])] = 1
                np_this_genotype[idx_subject, :] = list_allelType
            if FilterInLoading(np_this_genotype, np_phenotype) == 0:
                continue
            for idx_subject in range(0, int_num_phenotype):
                list_allelType = [0, 0, 0]
                list_allelType[np.argmax(list_thisSnp[idx_subject * 3 + 5 : idx_subject * 3 + 8])] = 1
                list_genotype[idx_subject].extend(list_allelType)
            list_genotype_rsid.append(list_thisSnp[1] + "_AA")
            list_genotype_rsid.append(list_thisSnp[1] + "_AB")
            list_genotype_rsid.append(list_thisSnp[1] + "_BB")
    np_genotype = np.array(list_genotype, dtype=np.int8)
    np_genotype_rsid = np.array(list_genotype_rsid)
    
    if np_genotype_rsid.shape[0] == 0:
        return 0.0
    
    #-------------------------
    # preprocess data
    #-------------------------    
    ### generate interaction terms
    np_genotype_rsid, np_genotype = FeatureEncoderLasso(np_genotype_rsid, np_genotype, np_phenotype, 3)
    
    #-------------------------
    # select feature
    #-------------------------    
    ### random lasso feature selection
    np_randWeight = np.array(RandomizedLassoRegression(np_genotype, np_phenotype[:, -1].astype(float)))
    np_selectedIdx = np.array([x >= 0.1 for x in np_randWeight])
    np_randWeight = np_randWeight[np_selectedIdx]
    np_genotype = np_genotype[:, np_selectedIdx]
    np_genotype_rsid = np_genotype_rsid[np_selectedIdx]
    if np_genotype_rsid.shape[0] == 0:
        return 0.0
    
    #-------------------------
    # build model
    #-------------------------
    float_AVG_S_P, np_weight = LassoRegressionCV(np_genotype, np_phenotype[:, -1].astype(float), int_kOfKFold, int_nJobs)
    if float_AVG_S_P == 0.0:
        return 0.0
    
    ### filter out zero-weight features
    np_selectedIdx = np.array([x != 0.0 for x in np_weight])
    np_weight = np_weight[np_selectedIdx]
    np_genotype = np_genotype[:, np_selectedIdx]
    np_genotype_rsid = np_genotype_rsid[np_selectedIdx]
    if np_genotype_rsid.shape[0] == 0:
        return 0.0
    
    #-------------------------
    # analyze result
    #-------------------------
    ### calculate student t-test p-value
    np_fRegression = -np.log10(f_regression(np_genotype.astype(int), np_phenotype[:, -1].astype(float))[1])
        
    ### calculate genotype frequency
    np_genotypeFreq = np.sum(np_genotype, axis=0).astype(float) / np_genotype.shape[0]
    
    #-------------------------
    # output results
    #-------------------------
    ### output statistics of features
    with open(os.path.join(str_outputFilePath, os.path.basename(str_inputFileName_genotype).split("_")[0] + "_Result.csv"), "w") as file_outputFile:
        file_outputFile.writelines("rsID,weight,student-t-test_log_p-value,genotype_frequency" + "\n")
        for idx_feature in range(0, np_genotype_rsid.shape[0]):
            file_outputFile.writelines(str(np_genotype_rsid[idx_feature,]) + "," + str(np_weight[idx_feature,]) + "," + str(np_fRegression[idx_feature,]) + "," + str(np_genotypeFreq[idx_feature]) + "\n")
    
    ### output feature
    with open(os.path.join(str_outputFilePath, os.path.basename(str_inputFileName_genotype).split("_")[0] + "_Feature.csv"), "w") as file_outputFile:
        file_outputFile.writelines(",".join(np_genotype_rsid) + "\n")
        for idx_subject in range(0, np_genotype.shape[0]):
            file_outputFile.writelines(",".join(np_genotype[idx_subject, :].astype(str)) + "\n")
    
    return float_AVG_S_P

def BatchSingleGeneEpistasisLasso(str_inputFilePath_genotype, str_inputFileName_phenotype, str_outputFilePath = "", int_kOfKFold = 2, int_nJobs = mp.cpu_count()):
    """

    Batch running for the single gene workflow.

    Args:
        str_inputFilePath_genotype (str): File path of input genotype data
        str_inputFileName_phenotype (str): File name of input phenotype data
        str_outputFilePath (str): File path of output file
        int_kOfKFold (int): The k for k-fold cross validation (default: 2)
        int_nJobs (int): The number of thread (default: 1)

    Returns:
        - Expected Success Response::

            "step4: Detect single gene epistasis. DONE!"
    
    """
    
    ### set default output path
    if str_outputFilePath == "":
        str_outputFilePath = os.path.abspath(os.path.join(str_inputFilePath_genotype, os.pardir)) + "/singleGeneResult/"
    ### if output folder doesn't exist then create it
    if not os.path.exists(str_outputFilePath):
        os.makedirs(str_outputFilePath)
    
    ### scan all of the gen file in path
    list_genotypeFileName = []
    for str_fileName in os.listdir(str_inputFilePath_genotype):
        if ".gen" in str_fileName:
            list_genotypeFileName.append(str_fileName)
    
    ### batch PolyLassoRegression
    ### inital multiprocessing pool
    mp_pool = mp.Pool(int_nJobs)

    ### apply pool on the function that need be parallelizing
    dict_result = {}
    for int_count_gene, float_AVG_S_P in enumerate(mp_pool.starmap(SingleGeneEpistasisLasso, [(os.path.join(str_inputFilePath_genotype, gene), str_inputFileName_phenotype, str_outputFilePath, int_kOfKFold, int_nJobs) for gene in list_genotypeFileName]), 0):
        if list_genotypeFileName[int_count_gene] not in dict_result:
            dict_result[list_genotypeFileName[int_count_gene]] = float_AVG_S_P
        str_print = "step4: Processing: " + "{0:.2f}".format(float(int_count_gene) / len(list_genotypeFileName) * 100) + "% - " + list_genotypeFileName[int_count_gene] + ": " + "\t\t"
        sys.stdout.write('%s\r' % str_print)
        sys.stdout.flush()

    mp_pool.close()

    ### output result
    with open(str_outputFilePath + "All_Lasso_k" + str(int_kOfKFold) + ".csv", "w") as file_outputFile:
        file_outputFile.writelines("GeneSymbol,AVG_S_P" + "\n")
        for key, value in dict_result.items():
            file_outputFile.writelines(key.split("_")[0] + "," + str(value) + "\n")

    '''
    ### batch PolyLassoRegression
    int_count_gene = 0
    with open(str_outputFilePath + "All_Lasso_k" + str(int_kOfKFold) + ".csv", "w") as file_outputFile:
        file_outputFile.writelines("GeneSymbol,AVG_S_P" + "\n")
        for item in list_genotypeFileName:
            int_count_gene = int_count_gene + 1
            str_genotypeFileName = os.path.join(str_inputFilePath_genotype, item)
            float_AVG_S_P = SingleGeneEpistasisLasso(str_genotypeFileName, str_inputFileName_phenotype, str_outputFilePath, int_kOfKFold, int_nJobs)
            file_outputFile.writelines(item.split("_")[0] + "," + str(float_AVG_S_P) + "\n")
            str_print = "step4: Processing: " + "{0:.2f}".format(float(int_count_gene) / len(list_genotypeFileName) * 100) + "% - " + item + ": " + str(float_AVG_S_P) + "\t\t"
            sys.stdout.write('%s\r' % str_print)
            sys.stdout.flush()
    '''

    print("step4: Detect single gene epistasis. DONE! \t\t\t\t")


# -*- coding: utf-8 -*-
"""
Created on Apr 2019

@author: Chester (Yu-Chuan Chang)
"""

""""""""""""""""""""""""""""""
# import libraries
""""""""""""""""""""""""""""""
import argparse
import os
import genepi
import numpy as np

""""""""""""""""""""""""""""""
# define functions 
""""""""""""""""""""""""""""""
def ArgumentsParser():
    ### define arguments
    str_description = ''
    'This script is a tool for extracting single gene sequence from whole genome reference'
    parser = argparse.ArgumentParser(prog='creategeneref', description=str_description)
    
    ### define arguments for I/O
    parser.add_argument("-i", required=True, help="filename of the input txt file")
    parser.add_argument("-r", required=True, help="filename of the reference genome")    
    parser.add_argument("-o", required=False, help="output file path")
    
    ### define other optional arguments    
    parser.add_argument("-b", required=False, default="hg19", choices=["hg19", "hg38"], help="human genome build")
    
    return parser

""""""""""""""""""""""""""""""
# main function
""""""""""""""""""""""""""""""
def main(args=None):
    ### obtain arguments from argument parser
    args = ArgumentsParser().parse_args(args)

    ### get arguments for I/O
    str_inputFileName_in = args.i
    str_inputFileName_ref = args.r
    str_outputFilePath = ""
    if args.o is not None:
        str_outputFilePath = args.o
    else:
        str_outputFilePath = os.path.dirname(str_inputFileName_ref)
        
    ### get gene symbol from input file
    list_gene = []
    with open(str_inputFileName_in, "r") as file_inputFile:
        for line in file_inputFile:
            list_gene.append(line.strip())
    
    ### switch genome build
    if args.b == "hg38":
        genepi.DownloadUCSCDB(str_hgbuild=args.b)
    
    ### load UCSC Genome Database
    list_UCSCGenomeDatabase = []
    with open(os.path.join(os.path.dirname(genepi.__file__), "UCSCGenomeDatabase.txt"), "r") as file_inputFile:
        for line in file_inputFile:
            list_UCSCGenomeDatabase.append(line.strip().split(","))
    np_UCSCGenomeDatabase = np.array(list_UCSCGenomeDatabase)
    
    ### get start & end of gene
    np_gene = np_UCSCGenomeDatabase[np.where(np_UCSCGenomeDatabase[:, 4] == list_gene[0]), :].ravel()
    if np_gene[3] == "+":
        np_gene[1] = int(np_gene[1]) + 1000
    else:
        np_gene[2] = int(np_gene[2]) - 1000
    
    ### output bed file
    with open(os.path.join(os.path.dirname(os.path.abspath(__file__)), "temp.bed"), "w") as file_outputFile:
        file_outputFile.writelines("chr" + str(np_gene[0]).replace("chr", "") + "\t" + str(np_gene[1]) + "\t" + str(np_gene[2]))
    
    ### use bedtools to extract single gene sequence
    str_command = "bedtools getfasta -fi " + str_inputFileName_ref + " -bed " + os.path.join(os.path.dirname(os.path.abspath(__file__)), "temp.bed")
    list_fasta = str(os.popen(str_command).read()).strip().split("\n")
    list_fasta[0] = ">" + str(list_gene[0])
    
    ### output single gene fasta
    with open(os.path.join(str_outputFilePath, str(np_gene[4]) + ".fasta"), "w") as file_outputFile:
        for line in list_fasta:
            file_outputFile.writelines(line + "\n")
    
    ### use samtools to generate index
    str_command = "samtools faidx " + os.path.join(str_outputFilePath, str(np_gene[4]) + ".fasta")
    os.system(str_command)
    
    ### use GATK to generate dict
    str_command = "python3 /opt/gatk/gatk CreateSequenceDictionary -R " + os.path.join(str_outputFilePath, str(np_gene[4]) + ".fasta") + " -O " + os.path.join(str_outputFilePath, str(np_gene[4]) + ".dict")
    os.system(str_command)
    
if __name__ == "__main__":
    main()


# -*- coding: utf-8 -*-
"""
Created on Feb 2019

@author: Chester (Yu-Chuan Chang)
"""

def warn(*args, **kwargs):
    pass
import warnings
warnings.warn = warn

""""""""""""""""""""""""""""""
# import libraries
""""""""""""""""""""""""""""""
import sys

""""""""""""""""""""""""""""""
# main function
""""""""""""""""""""""""""""""
def CorrectMissingID(str_inputFileName_genotype):
    ### count lines of input files
    int_num_genotype = sum(1 for line in open(str_inputFileName_genotype))
    
    ### read .gen file
    with open(str_inputFileName_genotype, "r") as file_inputFile:
        with open(str_inputFileName_genotype.replace(".gen", "_addID.gen"), "w") as file_outputFile:
            int_count_snp = 0
            for line in file_inputFile:
                list_thisSnp = line.strip().split(" ")
                ### if current snp have no ID, assign temporary ID (chromosome + position) to it
                if list_thisSnp[1] == ".":
                    list_thisSnp[1] = 'C' + str(list_thisSnp[0]) + 'P' + str(list_thisSnp[2])
                    file_outputFile.writelines(' '.join(list_thisSnp) + "\n")
                else:
                    file_outputFile.writelines(line + "\n")
                
                ### show progress
                int_count_snp = int_count_snp + 1
                str_print = "step0: Processing: " + "{0:.2f}".format(float(int_count_snp) / int_num_genotype * 100) + "\t\t"
                sys.stdout.write('%s\r' % str_print)
                sys.stdout.flush()
    
    print("step0: Correct Missingn ID DONE! \t\t\t\t")


# -*- coding: utf-8 -*-
"""
Created on Feb 2018

@author: Chester (Yu-Chuan Chang)
"""

""""""""""""""""""""""""""""""
# import libraries
""""""""""""""""""""""""""""""
import os
import pymysql

""""""""""""""""""""""""""""""
# main function
""""""""""""""""""""""""""""""
# database schema: http://hgdownload.cse.ucsc.edu/goldenpath/hg19/database/
# human genome build could be: hg18, hg19, hg38, etc.
def DownloadUCSCDB(str_outputFilePath = os.path.dirname(os.path.abspath(__file__)), str_hgbuild = "hg19"):
    """

    To obtain the gene information such as official gene symbols and genomic coordinates, this function is for retrieving kgXref and knownGene data table from the UCSC human genome annotation database

    Args:
        str_outputFilePath (str): File path of output database
        str_hgbuild (str): Genome build (eg. "hg19")

    Returns:
        - Expected Success Response::

            "step1: Down load UCSC Database. DONE!"
    
    """

    ### create connection
    conv = {pymysql.constants.FIELD_TYPE.LONG: int}
    conn = pymysql.Connect(host = "genome-mysql.cse.ucsc.edu", user = "genome", passwd = "",db = str_hgbuild,  local_infile = 1, conv = conv)
    
    ### execute sql command
    str_sqlCommand = "SELECT chr, CASE WHEN strand='+' THEN txStart-1000 ELSE txStart END AS txStart, CASE WHEN strand='-' THEN txEnd+1000 ELSE txEnd END AS txEnd, strand, geneSymbol FROM "
    str_sqlCommand = str_sqlCommand + "(SELECT REPLACE(chr, 'chr', '') AS chr, txStart, txEnd, strand, geneSymbol, MAX(ABS(txEnd-txStart)) FROM ( "
    str_sqlCommand = str_sqlCommand + "SELECT knownGene.chrom AS chr, knownGene.txStart AS txStart, knownGene.txEnd AS txEnd, knownGene.strand AS strand, kgXref.geneSymbol AS geneSymbol FROM kgXref INNER JOIN knownGene ON kgXref.kgID=knownGene.name WHERE LEFT(kgXref.mRNA, 2) IN ('NR', 'NM')) AS L1 "
    str_sqlCommand = str_sqlCommand + "GROUP BY geneSymbol) AS L2 WHERE LEFT(chr, 1) NOT IN ('X', 'Y', 'M', 'U') AND chr NOT LIKE '%\_%' ORDER BY CAST(chr AS UNSIGNED), txStart"
    cur = conn.cursor()
    cur.execute(str_sqlCommand)
    db_result = cur.fetchall()
    cur.close()
    conn.close()

    ### output database
    with open(str_outputFilePath + "/UCSCGenomeDatabase.txt", "w") as file_outputFile:
        for item in db_result:
            file_outputFile.writelines(",".join(item) + "\n")
    
    print("step1: Down load UCSC Database. DONE!")



# -*- coding: utf-8 -*-
"""
Created on Feb 2018

@author: Chester (Yu-Chuan Chang)
"""

""""""""""""""""""""""""""""""
# import libraries
""""""""""""""""""""""""""""""
import os
import warnings
warnings.filterwarnings('ignore')
# ignore all warnings
warnings.simplefilter("ignore")
os.environ["PYTHONWARNINGS"] = "ignore"

import os
import numpy as np
np.seterr(divide='ignore', invalid='ignore')
from sklearn.feature_selection import f_regression
from sklearn import linear_model
from scipy.sparse import coo_matrix
from sklearn.utils import shuffle
from sklearn.model_selection import KFold
from sklearn.model_selection import GridSearchCV
from sklearn.externals import joblib
import scipy.stats as stats

from genepi.step4_singleGeneEpistasis_Lasso import RandomizedLassoRegression
from genepi.step4_singleGeneEpistasis_Lasso import LassoRegressionCV
from genepi.step4_singleGeneEpistasis_Lasso import FeatureEncoderLasso

""""""""""""""""""""""""""""""
# define functions 
""""""""""""""""""""""""""""""
def LassoRegression(np_X, np_y, int_nJobs = 4):
    """

    Implementation of the L1-regularized Lasso regression with k-fold cross validation.

    Args:
        np_X (ndarray): 2D array containing genotype data with `int8` type
        np_y (ndarray): 2D array containing phenotype data with `float` type
        int_nJobs (int): The number of thread (default: 4)

    Returns:
        (float): float_AVG_S_P
        
            The average of the Peason's and Spearman's correlation of the model
    
    """

    X = np_X
    y = np_y
    
    list_target = []
    list_predict = []
    
    alpha = np.logspace(-10, 10, 200)
    parameters = [{'alpha':alpha}]
    kf_estimator = KFold(n_splits=2)
    estimator_lasso = linear_model.Lasso(max_iter=1000)
    estimator_grid = GridSearchCV(estimator_lasso, parameters, scoring='neg_mean_squared_error', n_jobs=int_nJobs, cv=kf_estimator)
    estimator_grid.fit(X, y)
    list_label = estimator_grid.best_estimator_.predict(X)
    for idx_y, idx_label in zip(list(y), list_label):
        list_target.append(float(idx_y))
        list_predict.append(idx_label)
    float_pearson = stats.stats.pearsonr(list_target, list_predict)[0]
    float_spearman = stats.stats.spearmanr(list_target, list_predict)[0]
    
    return (float_pearson + float_spearman) / 2

def RegressorModelPersistence(np_X, np_y, str_outputFilePath = "", int_nJobs = 4):
    """

    Dumping regressor for model persistence

    Args:
        np_X (ndarray): 2D array containing genotype data with `int8` type
        np_y (ndarray): 2D array containing phenotype data with `float` type
        str_outputFilePath (str): File path of output file
        int_nJobs (int): The number of thread (default: 4)

    Returns:
        None
    
    """

    X = np_X
    y = np_y
    X_sparse = coo_matrix(X)
    X, X_sparse, y = shuffle(X, X_sparse, y, random_state=0)
    
    alpha = np.logspace(-10, 10, 200)
    parameters = [{'alpha':alpha}]
    kf_estimator = KFold(n_splits=2)
    estimator_lasso = linear_model.Lasso(max_iter=1000)
    estimator_grid = GridSearchCV(estimator_lasso, parameters, scoring='neg_mean_squared_error', n_jobs=int_nJobs, cv=kf_estimator)
    estimator_grid.fit(X, y)
    
    joblib.dump(estimator_grid.best_estimator_, os.path.join(str_outputFilePath, "Regressor.pkl"))

""""""""""""""""""""""""""""""
# main function
""""""""""""""""""""""""""""""
def CrossGeneEpistasisLasso(str_inputFilePath_feature, str_inputFileName_phenotype, str_inputFileName_score = "", str_outputFilePath = "", int_kOfKFold = 2, int_nJobs = 4):
    """

    A workflow to model a cross gene epistasis containing two-element combinatorial encoding, stability selection, filtering low quality varaint and  L1-regularized Lasso regression with k-fold cross validation.

    Args:
        str_inputFilePath_feature (str): File path of input feature files from stage 1 - singleGeneEpistasis
        str_inputFileName_phenotype (str): File name of input phenotype data
        str_inputFileName_score (str): File name of input score file from stage 1 - singleGeneEpistasis
        str_outputFilePath (str): File path of output file
        int_kOfKFold (int): The k for k-fold cross validation (default: 2)
        int_nJobs (int): The number of thread (default: 4)

    Returns:
        (tuple): tuple containing:

            - float_AVG_S_P_train (float): The average of the Peason's and Spearman's correlation of the model for training set
            - float_AVG_S_P_test (float): The average of the Peason's and Spearman's correlation of the model for testing set

        - Expected Success Response::

            "step5: Detect cross gene epistasis. DONE!"
    
    """
    
    ### set default output path
    if str_outputFilePath == "":
        str_outputFilePath = os.path.abspath(os.path.join(str_inputFilePath_feature, os.pardir)) + "/crossGeneResult/"
    ### if output folder doesn't exist then create it
    if not os.path.exists(str_outputFilePath):
        os.makedirs(str_outputFilePath)
    
    ### set default score file name
    if str_inputFileName_score == "":
        for str_fileName in os.listdir(str_inputFilePath_feature):
            if str_fileName.startswith("All_Lasso"):
                str_inputFileName_score = os.path.join(str_inputFilePath_feature, str_fileName)
    
    #-------------------------
    # load data
    #-------------------------
    ### scan score file and exclude useless genes
    dict_score = {}
    with open(str_inputFileName_score, "r") as file_inputFile:
        file_inputFile.readline()
        for line in file_inputFile:
            list_thisScore = line.strip().split(",")
            if list_thisScore[1] == "MemErr" or float(list_thisScore[1]) == 0.0:
                pass
            else:
                dict_score[list_thisScore[0]] = float(list_thisScore[1])
    
    ### get all the file names of feature file
    list_featureFileName = []
    for str_fileName in os.listdir(str_inputFilePath_feature):
        if "Feature.csv" in str_fileName:
            list_featureFileName.append(str_fileName)
    
    ### get all selected snp ids
    list_genotype_rsid = []
    for item in list_featureFileName:
        with open(os.path.join(str_inputFilePath_feature, item), "r") as file_inputFile:
            ### grep the header
            list_rsids = file_inputFile.readline().strip().split(",")
            for rsid in list_rsids:
                list_genotype_rsid.append(rsid)
    np_genotype_rsid = np.array(list_genotype_rsid)
    
    ### count lines of input files
    int_num_genotype = len(np_genotype_rsid)
    int_num_phenotype = sum(1 for line in open(str_inputFileName_phenotype))
    
    ### get phenotype file
    list_phenotype = []
    with open(str_inputFileName_phenotype, 'r') as file_inputFile:
        for line in file_inputFile:
            list_phenotype.append(line.strip().split(","))
    np_phenotype = np.array(list_phenotype, dtype=np.float)
    del list_phenotype
    
    ### get genotype file
    ### declare a dictionary for mapping snp and gene
    dict_geneMap ={}
    idx_genotype_rsid = 0
    np_genotype = np.empty([int_num_phenotype, int_num_genotype], dtype='int8')
    for item in list_featureFileName:
        with open(os.path.join(str_inputFilePath_feature, item), "r") as file_inputFile:
            ### grep feature from header of feature file
            list_rsids = file_inputFile.readline().strip().split(",")
            for rsid in list_rsids:
                ### key: rsIDs of a feature; value: gene symbol
                dict_geneMap[rsid] = item.split("_")[0]
            idx_phenotype = 0
            ### read feaure and write into np_genotype
            for line in file_inputFile:
                np_genotype[idx_phenotype, idx_genotype_rsid:idx_genotype_rsid + len(list_rsids)] = np.array([float(x) for x in line.strip().split(",")], dtype='int')
                idx_phenotype = idx_phenotype + 1
            idx_genotype_rsid = idx_genotype_rsid + len(list_rsids)
    
    #-------------------------
    # preprocess data
    #-------------------------
    ### f regression feature selection
    np_fRegression = -np.log10(f_regression(np_genotype.astype(int), np_phenotype[:, -1].astype(float))[1])
    np_selectedIdx = np.array([x > 5 for x in np_fRegression])
    np_genotype = np_genotype[:, np_selectedIdx]
    np_genotype_rsid = np_genotype_rsid[np_selectedIdx]
    if np_genotype_rsid.shape[0] == 0:
        return 0.0

    ### select degree 1 feature
    np_genotype_rsid_degree = np.array([str(x).count('*') + 1 for x in np_genotype_rsid])
    np_selectedIdx = np.array([x == 1 for x in np_genotype_rsid_degree])
    np_genotype_degree1 = np_genotype[:, np_selectedIdx]
    np_genotype_degree1_rsid = np_genotype_rsid[np_selectedIdx]
    
    ### remove redundant polynomial features
    np_genotype_degree1, np_selectedIdx = np.unique(np_genotype_degree1, axis=1, return_index=True)
    np_genotype_degree1_rsid = np_genotype_degree1_rsid[np_selectedIdx]
    
    ### generate cross gene interations
    np_genotype_crossGene_rsid, np_genotype_crossGene = FeatureEncoderLasso(np_genotype_degree1_rsid, np_genotype_degree1, np_phenotype, 1)
    
    ### remove degree 1 feature from dataset
    np_selectedIdx = np.array([x != 1 for x in np_genotype_rsid_degree])
    np_genotype = np_genotype[:, np_selectedIdx]
    np_genotype_rsid = np_genotype_rsid[np_selectedIdx]
    
    ### concatenate cross gene interations
    if np_genotype_degree1.shape[1] > 0:
        np_genotype = np.concatenate((np_genotype, np_genotype_crossGene), axis=1)
        np_genotype_rsid = np.concatenate((np_genotype_rsid, np_genotype_crossGene_rsid))
    
    #-------------------------
    # select feature
    #-------------------------
    ### random lasso feature selection
    np_randWeight = np.array(RandomizedLassoRegression(np_genotype, np_phenotype[:, -1].astype(float)))
    np_selectedIdx = np.array([x >= 0.1 for x in np_randWeight])
    np_randWeight = np_randWeight[np_selectedIdx]
    np_genotype = np_genotype[:, np_selectedIdx]
    np_genotype_rsid = np_genotype_rsid[np_selectedIdx]
    if np_genotype_rsid.shape[0] == 0:
        return 0.0
    
    #-------------------------
    # build model
    #-------------------------
    float_AVG_S_P_test, np_weight = LassoRegressionCV(np_genotype, np_phenotype[:, -1].astype(float), int_kOfKFold, int_nJobs)
    float_AVG_S_P_train = LassoRegression(np_genotype, np_phenotype[:, -1].astype(float), int_nJobs)
    
    ### filter out zero-weight features
    np_selectedIdx = np.array([x != 0.0 for x in np_weight])
    np_weight = np_weight[np_selectedIdx]
    np_genotype = np_genotype[:, np_selectedIdx]
    np_genotype_rsid = np_genotype_rsid[np_selectedIdx]
    if np_genotype_rsid.shape[0] == 0:
        return 0.0
    
    #-------------------------
    # analyze result
    #-------------------------
    ### calculate student t-test p-value
    np_fRegression = -np.log10(f_regression(np_genotype.astype(int), np_phenotype[:, -1].astype(float))[1])
        
    ### calculate genotype frequency
    np_genotypeFreq = np.sum(np_genotype, axis=0).astype(float) / np_genotype.shape[0]
    
    #-------------------------
    # output results
    #-------------------------
    ### output statistics of features
    with open(os.path.join(str_outputFilePath, "Result.csv"), "w") as file_outputFile:
        file_outputFile.writelines("rsid,weight,student-t-test_log_p-value,genotype_frequency,geneSymbol,singleGeneScore" + "\n")
        for idx_feature in range(0, np_genotype_rsid.shape[0]):
            ### if this feature is single gene epistasis
            if np_genotype_rsid[idx_feature,] in dict_geneMap.keys():
                str_thisOutput = str(np_genotype_rsid[idx_feature,]) + "," + str(np_weight[idx_feature,]) + "," + str(np_fRegression[idx_feature,]) + "," + str(np_genotypeFreq[idx_feature]) + "," + str(dict_geneMap[np_genotype_rsid[idx_feature,]]).split("@")[0] + "," + str(dict_score[dict_geneMap[np_genotype_rsid[idx_feature,]]]) + "\n"
                file_outputFile.writelines(str_thisOutput)
            ### else this feature is cross gene epistasis
            else:
                str_thisOutput = str(np_genotype_rsid[idx_feature,]) + "," + str(np_weight[idx_feature,]) + "," + str(np_fRegression[idx_feature,]) + "," + str(np_genotypeFreq[idx_feature]) + "," + str(dict_geneMap[np_genotype_rsid[idx_feature,].split("*")[0]]).split("@")[0] + "*" + str(dict_geneMap[np_genotype_rsid[idx_feature,].split("*")[1]]).split("@")[0] + ", " + "\n"
                file_outputFile.writelines(str_thisOutput)

    ### output feature
    with open(os.path.join(str_outputFilePath, "Feature.csv"), "w") as file_outputFile:
        file_outputFile.writelines(",".join(np_genotype_rsid) + "\n")
        for idx_subject in range(0, np_genotype.shape[0]):
            file_outputFile.writelines(",".join(np_genotype[idx_subject, :].astype(str)) + "\n")

    #-------------------------
    # dump persistent model
    #-------------------------
    RegressorModelPersistence(np_genotype, np_phenotype[:, -1].astype(int), str_outputFilePath, int_nJobs)

    print("step5: Detect cross gene epistasis. DONE! (Training score:" + "{0:.2f}".format(float_AVG_S_P_train) + "; " + str(int_kOfKFold) + "-fold Test Score:" + "{0:.2f}".format(float_AVG_S_P_test) + ")")
    
    return float_AVG_S_P_train, float_AVG_S_P_test
    



# -*- coding: utf-8 -*-
"""
Created on Feb 2018

@author: Chester (Yu-Chuan Chang)
"""

""""""""""""""""""""""""""""""
# import libraries
""""""""""""""""""""""""""""""
from .step0_correctMissingID import CorrectMissingID
from .step1_downloadUCSCDB import DownloadUCSCDB
from .step2_estimateLD import EstimateLDBlock
from .step3_splitByGene import SplitByGene
from .step4_singleGeneEpistasis_Logistic import SingleGeneEpistasisLogistic
from .step4_singleGeneEpistasis_Logistic import BatchSingleGeneEpistasisLogistic
from .step4_singleGeneEpistasis_Logistic import RandomizedLogisticRegression
from .step4_singleGeneEpistasis_Logistic import LogisticRegressionL1CV
from .step4_singleGeneEpistasis_Logistic import FeatureEncoderLogistic
from .step4_singleGeneEpistasis_Lasso import SingleGeneEpistasisLasso
from .step4_singleGeneEpistasis_Lasso import BatchSingleGeneEpistasisLasso
from .step4_singleGeneEpistasis_Lasso import RandomizedLassoRegression
from .step4_singleGeneEpistasis_Lasso import LassoRegressionCV
from .step4_singleGeneEpistasis_Lasso import FeatureEncoderLasso
from .step5_crossGeneEpistasis_Logistic import CrossGeneEpistasisLogistic
from .step5_crossGeneEpistasis_Logistic import LogisticRegressionL1
from .step5_crossGeneEpistasis_Lasso import CrossGeneEpistasisLasso
from .step5_crossGeneEpistasis_Lasso import LassoRegression
from .step6_ensembleWithCovariates import EnsembleWithCovariatesClassifier
from .step6_ensembleWithCovariates import EnsembleWithCovariatesRegressor
from .step7_validateByIsolatedData import SplittingDataAsIsolatedData
from .step7_validateByIsolatedData import ValidateByIsolatedDataClassifier
from .step7_validateByIsolatedData import ValidateByIsolatedDataRegressor
from .step7_validateByIsolatedData import ValidateByIsolatedDataCovariateClassifier
from .step7_validateByIsolatedData import ValidateByIsolatedDataCovariateRegressor


# -*- coding: utf-8 -*-
"""
Created on Feb 2018

@author: Chester (Yu-Chuan Chang)
"""

def warn(*args, **kwargs):
    pass
import warnings
warnings.warn = warn

""""""""""""""""""""""""""""""
# import libraries
""""""""""""""""""""""""""""""
import os
import numpy as np
from sklearn.externals import joblib
import sklearn.metrics as skMetric
import scipy.stats as stats

""""""""""""""""""""""""""""""
# define functions 
""""""""""""""""""""""""""""""
def SplittingDataAsIsolatedData(str_inputFileName_genotype, str_inputFileName_phenotype, str_outputFilePath = "", int_randomState = 0):
    ### set path of output file
    if str_outputFilePath == "":
        str_outputFilePath = os.path.dirname(str_inputFileName_genotype)    
    
    #-------------------------
    # load data
    #-------------------------
    ### count lines of input files
    int_num_genotype = sum(1 for line in open(str_inputFileName_genotype))
    int_num_phenotype = sum(1 for line in open(str_inputFileName_phenotype))
    
    ### get phenotype file
    list_phenotype = []
    with open(str_inputFileName_phenotype, 'r') as file_inputFile:
        for line in file_inputFile:
            list_phenotype.append(line.strip().split(","))
    np_phenotype = np.array(list_phenotype)
    del list_phenotype
    
    ### get genotype file
    list_genotype_info = []
    list_genotype = []
    with open(str_inputFileName_genotype, 'r') as file_inputFile:
        for line in file_inputFile:
            list_thisSnp = line.strip().split(" ")
            list_genotype_info.append(list_thisSnp[:5])
            list_genotype.append(list_thisSnp[5:])
    np_genotype_info = np.array(list_genotype_info)
    del list_genotype_info
    np_genotype = np.array(list_genotype)
    del list_genotype
    
    ### set random state
    random = np.random.RandomState(int_randomState)
    ### random sample
    np_random = random.choice(int_num_phenotype, int(int_num_phenotype/2), replace=False)
    np_random_complement = np.ones(int_num_phenotype, dtype=bool)
    np_random_complement[np_random] = False
    
    #-------------------------
    # output results
    #-------------------------
    ### output phenotype files
    with open(os.path.join(str_outputFilePath, os.path.basename(str_inputFileName_phenotype).replace(".csv", "_subset_1.csv")), "w") as file_outputFile:
        np_phenotype_selected = np_phenotype[np_random, :]
        for idx_phenotype in range(np_phenotype_selected.shape[0]):
            str_line = ",".join(np_phenotype_selected[idx_phenotype, :])
            file_outputFile.writelines(str_line + "\n")
    with open(os.path.join(str_outputFilePath, os.path.basename(str_inputFileName_phenotype).replace(".csv", "_subset_2.csv")), "w") as file_outputFile:
        np_phenotype_selected = np_phenotype[np_random_complement, :]
        for idx_phenotype in range(np_phenotype_selected.shape[0]):
            str_line = ",".join(np_phenotype_selected[idx_phenotype, :])
            file_outputFile.writelines(str_line + "\n")
    
    ### output genotype files
    with open(os.path.join(str_outputFilePath, os.path.basename(str_inputFileName_genotype).replace(".gen", "_subset_1.gen")), "w") as file_outputFile:
        np_genotype_selected = np_genotype[:, np.array([[x * 3, x * 3 + 1, x * 3 + 2] for x in np_random]).ravel()]
        for idx_genotype in range(int_num_genotype):
            str_line = " ".join(np_genotype_info[idx_genotype, :]) + " " + " ".join(np_genotype_selected[idx_genotype, :])
            file_outputFile.writelines(str_line + "\n")
    with open(os.path.join(str_outputFilePath, os.path.basename(str_inputFileName_genotype).replace(".gen", "_subset_2.gen")), "w") as file_outputFile:
        np_genotype_selected = np_genotype[:, np.array([[x, x, x] for x in np_random_complement]).ravel()]
        for idx_genotype in range(int_num_genotype):
            str_line = " ".join(np_genotype_info[idx_genotype, :]) + " " + " ".join(np_genotype_selected[idx_genotype, :])
            file_outputFile.writelines(str_line + "\n")

def IsolatedDataFeatureGenerator(str_inputFileName_feature, str_inputFileName_genotype, str_inputFileName_phenotype, str_outputFilePath = ""):
    ### set default output path
    if str_outputFilePath == "":
        str_outputFilePath = os.path.dirname(str_inputFileName_genotype) + "/isolatedValidation/"
    ### if output folder doesn't exist then create it
    if not os.path.exists(str_outputFilePath):
        os.makedirs(str_outputFilePath)

    ### get all selected snp ids
    list_feature_rsid_all = []
    with open(str_inputFileName_feature, "r") as file_inputFile:
        ### grep the header
        list_rsids = file_inputFile.readline().strip().split(",")
        for rsid in list_rsids:
            list_feature_rsid_all.append(rsid)
    ### get unique selected snp ids
    dict_feature_rsid_unique = {}
    for item in list_feature_rsid_all:
        for subitem in item.replace("_AA", "").replace("_AB", "").replace("_BB", "").split(" "):
            if subitem not in dict_feature_rsid_unique:
                dict_feature_rsid_unique[subitem] = 1
    
    ### extract selected snp from genotype file
    list_inputFile_genotype = []
    with open(str_inputFileName_genotype, 'r') as file_inputFile:
        for line in file_inputFile:
            list_thisSnp = line.strip().split(" ")
            if list_thisSnp[1] in dict_feature_rsid_unique:
                list_inputFile_genotype.append(line)
    
    ### count lines of input files
    int_num_genotype = len(list_inputFile_genotype)
    int_num_phenotype = sum(1 for line in open(str_inputFileName_phenotype))
    
    ### get phenotype file
    list_phenotype = []
    with open(str_inputFileName_phenotype, 'r') as file_inputFile:
        for line in file_inputFile:
            list_phenotype.append(line.strip().split(","))
    np_phenotype = np.array(list_phenotype)
    del list_phenotype
    
    list_genotype_rsid = []
    ### declare a numpy array for one-hot-encoded genotype
    np_genotype = np.empty([int_num_phenotype, int_num_genotype * 3], dtype='int')
    idx_snp = 0
    for line in list_inputFile_genotype:
        list_thisSnp = line.strip().split(" ")
        list_genotype_rsid.append(list_thisSnp[1] + "_AA")
        list_genotype_rsid.append(list_thisSnp[1] + "_AB")
        list_genotype_rsid.append(list_thisSnp[1] + "_BB")
        for idx_subject in range(0, int_num_phenotype):
            list_allelType = [0, 0, 0]
            list_allelType[np.argmax(list_thisSnp[idx_subject * 3 + 5 : idx_subject * 3 + 8])] = 1
            np_genotype[idx_subject, idx_snp * 3 : idx_snp * 3 + 3] = list_allelType
        idx_snp = idx_snp + 1
    np_genotype_rsid = np.array(list_genotype_rsid)
    
    ### generate feature
    np_feature = np.empty([int_num_phenotype, len(list_feature_rsid_all)], dtype='int')
    for idx_feature in range(len(list_feature_rsid_all)):
        list_feature_rsid = list_feature_rsid_all[idx_feature].split(" ")
        if len(list_feature_rsid) == 1:
            np_feature[:, idx_feature] = np_genotype[:, int(np.argwhere(np_genotype_rsid == list_feature_rsid[0]))]
        else:
            np_feature[:, idx_feature] = np.multiply(np_genotype[:, int(np.argwhere(np_genotype_rsid == list_feature_rsid[0]))], np_genotype[:, int(np.argwhere(np_genotype_rsid == list_feature_rsid[1]))])
    
    ### output feature
    with open(os.path.join(str_outputFilePath, "Feature.csv"), "w") as file_outputFile:
        file_outputFile.writelines(",".join(list_feature_rsid_all) + "\n")
        for idx_subject in range(int_num_phenotype):
            file_outputFile.writelines(",".join(np_feature[idx_subject, :].astype(str)) + "\n")
    
    return np_feature, np_phenotype

def ValidateByIsolatedDataClassifier(str_inputFileName_model, str_inputFileName_feature, str_inputFileName_genotype, str_inputFileName_phenotype, str_outputFilePath = ""):
    ### set default output path
    if str_outputFilePath == "":
        str_outputFilePath = os.path.dirname(str_inputFileName_genotype) + "/isolatedValidation/"
    ### if output folder doesn't exist then create it
    if not os.path.exists(str_outputFilePath):
        os.makedirs(str_outputFilePath)
    
    estimator = joblib.load(str_inputFileName_model)
    np_genotype, np_phenotype = IsolatedDataFeatureGenerator(str_inputFileName_feature, str_inputFileName_genotype, str_inputFileName_phenotype, str_outputFilePath)
    
    list_target = []
    list_predict = []
    list_label = estimator.predict(np_genotype)
    for idx_target, idx_prdict in zip(list(np_phenotype[:, -1].astype(int)), list_label):
        list_target.append(float(idx_target))
        list_predict.append(idx_prdict)
    float_f1Score = skMetric.f1_score(list_target, list_predict)
    
    print("step7: Validate by isolated data. DONE! (Test score:" + "{0:.2f}".format(float_f1Score) + ")")
    
    return float_f1Score

def ValidateByIsolatedDataRegressor(str_inputFileName_model, str_inputFileName_feature, str_inputFileName_genotype, str_inputFileName_phenotype, str_outputFilePath = ""):
    ### set default output path
    if str_outputFilePath == "":
        str_outputFilePath = os.path.dirname(str_inputFileName_genotype) + "/isolatedValidation/"
    ### if output folder doesn't exist then create it
    if not os.path.exists(str_outputFilePath):
        os.makedirs(str_outputFilePath)
    
    estimator = joblib.load(str_inputFileName_model)
    np_genotype, np_phenotype = IsolatedDataFeatureGenerator(str_inputFileName_feature, str_inputFileName_genotype, str_inputFileName_phenotype, str_outputFilePath)
    
    list_target = []
    list_predict = []
    list_label = estimator.predict(np_genotype)
    for idx_target, idx_prdict in zip(list(np_phenotype[:, -1].astype(int)), list_label):
        list_target.append(float(idx_target))
        list_predict.append(idx_prdict)
    float_pearson = stats.stats.pearsonr(list_target, list_predict)[0]
    float_spearman = stats.stats.spearmanr(list_target, list_predict)[0]
    float_AVG_S_P = (float_pearson + float_spearman) / 2
    
    print("step7: Validate by isolated data. DONE! (Test score:" + "{0:.2f}".format(float_AVG_S_P) + ")")

    return float_AVG_S_P

def ValidateByIsolatedDataCovariateClassifier(str_inputFileName_model, str_inputFileName_feature, str_inputFileName_genotype, str_inputFileName_phenotype, str_outputFilePath = ""):
    ### set default output path
    if str_outputFilePath == "":
        str_outputFilePath = os.path.dirname(str_inputFileName_genotype) + "/isolatedValidation/"
    ### if output folder doesn't exist then create it
    if not os.path.exists(str_outputFilePath):
        os.makedirs(str_outputFilePath)
    
    estimator = joblib.load(str_inputFileName_model)
    np_genotype, np_phenotype = IsolatedDataFeatureGenerator(str_inputFileName_feature, str_inputFileName_genotype, str_inputFileName_phenotype, str_outputFilePath)
    
    if np_phenotype.shape[1] < 2:
        print("step7: Error no other factors exist.")
        return 0.0
    
    ### concatenate genotype and other factors
    np_genotype = np.concatenate((np_genotype, np_phenotype[:, :-1]), axis=1).astype(float)
    
    list_target = []
    list_predict = []
    list_label = estimator.predict(np_genotype)
    for idx_target, idx_prdict in zip(list(np_phenotype[:, -1].astype(int)), list_label):
        list_target.append(float(idx_target))
        list_predict.append(idx_prdict)
    float_f1Score = skMetric.f1_score(list_target, list_predict)
    
    print("step7: Validate by isolated data(with covariates). DONE! (Test score:" + "{0:.2f}".format(float_f1Score) + ")")
    
    return float_f1Score

def ValidateByIsolatedDataCovariateRegressor(str_inputFileName_model, str_inputFileName_feature, str_inputFileName_genotype, str_inputFileName_phenotype, str_outputFilePath = ""):
    ### set default output path
    if str_outputFilePath == "":
        str_outputFilePath = os.path.dirname(str_inputFileName_genotype) + "/isolatedValidation/"
    ### if output folder doesn't exist then create it
    if not os.path.exists(str_outputFilePath):
        os.makedirs(str_outputFilePath)
    
    estimator = joblib.load(str_inputFileName_model)
    np_genotype, np_phenotype = IsolatedDataFeatureGenerator(str_inputFileName_feature, str_inputFileName_genotype, str_inputFileName_phenotype, str_outputFilePath)
    
    if np_phenotype.shape[1] < 2:
        print("step7: Error no other factors exist.")
        return 0.0
    
    ### concatenate genotype and other factors
    np_genotype = np.concatenate((np_genotype, np_phenotype[:, :-1]), axis=1).astype(float)
    
    list_target = []
    list_predict = []
    list_label = estimator.predict(np_genotype)
    for idx_target, idx_prdict in zip(list(np_phenotype[:, -1].astype(int)), list_label):
        list_target.append(float(idx_target))
        list_predict.append(idx_prdict)
    float_pearson = stats.stats.pearsonr(list_target, list_predict)[0]
    float_spearman = stats.stats.spearmanr(list_target, list_predict)[0]
    float_AVG_S_P = (float_pearson + float_spearman) / 2
    
    print("step7: Validate by isolated data(with covariates). DONE! (Test score:" + "{0:.2f}".format(float_AVG_S_P) + ")")

    return float_AVG_S_P


# -*- coding: utf-8 -*-
"""
Created on Feb 2018

@author: Chester (Yu-Chuan Chang)
"""

""""""""""""""""""""""""""""""
# import libraries
""""""""""""""""""""""""""""""
import os
import numpy as np

""""""""""""""""""""""""""""""
# main function
""""""""""""""""""""""""""""""
def SplitMegaGene(list_snpsOnGene, int_window, int_step, str_outputFilePath, str_outputFileName):
    """

    In order to extract genetic features for a gene, this function used the start and end positions of each gene from the local UCSC database to split the genetic features. Then, generate the .GEN files for each gene in the folder named snpSubsets.

    Args:
        list_snpsOnGene (list): A list contains SNPs on a gene
        int_window (int): The size of the sliding window
        int_step (int): The step of the sliding window
        str_outputFilePath (str): File path of output file
        str_outputFileName (str): File name of output file

    Returns:
        None
    
    """
    
    ### write to gen file if this gene is not mega gene
    int_total_window = int((len(list_snpsOnGene)-int_window)/int_step)
    if int_total_window <= 0:
        with open(os.path.join(str_outputFilePath, str_outputFileName + ".gen"), "w") as file_outputFile:
            for item in list_snpsOnGene:
                file_outputFile.writelines(item)

    ### write gen file of each window on current gene (output file name: geneSymbol_numOfSNPOnGene@windowNum.gen)
    else: 
        for idx_w in range(int_total_window):
            with open(os.path.join(str_outputFilePath, str_outputFileName.split("_")[0] + "@" + str(idx_w) + "_" + str_outputFileName.split("_")[1] + ".gen"), "w") as file_outputFile:
                for item in list_snpsOnGene[int_step*idx_w:int_step*idx_w+int_window]:
                    file_outputFile.writelines(item)
        
        ### write reminder SNPs to gen file
        with open(os.path.join(str_outputFilePath, str_outputFileName.split("_")[0] + "@" + str(int_total_window) + "_" + str_outputFileName.split("_")[1] + ".gen"), "w") as file_outputFile:
            for item in list_snpsOnGene[int_step*int_total_window:]:
                file_outputFile.writelines(item)
    
    return

def SplitByGene(str_inputFileName_genotype, str_inputFileName_UCSCDB = os.path.dirname(os.path.abspath(__file__)) + "/UCSCGenomeDatabase.txt", str_outputFilePath = ""):
    """

    In order to extract genetic features for a gene, this function used the start and end positions of each gene from the local UCSC database to split the genetic features. Then, generate the .GEN files for each gene in the folder named snpSubsets.

    Args:
        str_inputFileName_genotype (str): File name of input genotype data
        str_inputFileName_UCSCDB (str): File name of input genome regions
        str_outputFilePath (str): File path of output file

    Returns:
        - Expected Success Response::

            "step3: Split by gene. DONE!"
    
    Warnings:
        "Warning of step3: .gen file should be sorted by chromosome and position"
    
    """
    
    print("Warning of step3: .gen file should be sorted by chromosome and position")

    int_window = 1000
    int_step = 200
    
    ### set default output path
    if str_outputFilePath == "":
        str_outputFilePath = os.path.join(os.path.dirname(str_inputFileName_genotype), "snpSubsets")
    ### if output folder doesn't exist then create it
    if not os.path.exists(str_outputFilePath):
        os.makedirs(str_outputFilePath)
    
    ### load UCSC Genome Database
    list_UCSCGenomeDatabase = []
    with open(str_inputFileName_UCSCDB, "r") as file_inputFile:
        for line in file_inputFile:
            list_UCSCGenomeDatabase.append(line.strip().split(","))
    np_UCSCGenomeDatabase = np.array(list_UCSCGenomeDatabase)
    
    ### scan all snp
    with open(str_inputFileName_genotype, "r") as file_inputFile:
        idx_gene = 0
        list_snpsOnGene = []
        for line in file_inputFile:
            ### get information of each snp
            list_thisSnp = line.strip().split(" ")
            int_chromosome = int(list_thisSnp[0])
            int_position = int(list_thisSnp[2])

            ### current gene is in next chromosome
            if int_chromosome < int(np_UCSCGenomeDatabase[idx_gene, 0]):
                continue
            ### current snp of genotype data is in next chromosome
            elif int_chromosome > int(np_UCSCGenomeDatabase[idx_gene, 0]):
                if len(list_snpsOnGene) != 0:
                    #### write gen file of current gene (output file name: geneSymbol_numOfSNPOnGene.gen)
                    #str_outputFileName = str(np_UCSCGenomeDatabase[idx_gene, 4]) + "_" + str(len(list_snpsOnGene)) + ".gen"
                    #with open(os.path.join(str_outputFilePath, str_outputFileName), "w") as file_outputFile:
                    #    for item in list_snpsOnGene:
                    #        file_outputFile.writelines(item)
                    str_outputFileName = str(np_UCSCGenomeDatabase[idx_gene, 4]) + "_" + str(len(list_snpsOnGene))
                    SplitMegaGene(list_snpsOnGene, int_window, int_step, str_outputFilePath, str_outputFileName)
                list_snpsOnGene = []
                while int_chromosome > int(np_UCSCGenomeDatabase[idx_gene, 0]):
                    ### jump to next gene
                    idx_gene = idx_gene + 1
                    ### if no next gene then break
                    if idx_gene == np_UCSCGenomeDatabase.shape[0]:
                        break
                    ### current snp on next gene
                    if int(np_UCSCGenomeDatabase[idx_gene, 1]) <= int_position and int_position <= int(np_UCSCGenomeDatabase[idx_gene, 2]) and int_chromosome == int(np_UCSCGenomeDatabase[idx_gene, 0]):
                        list_snpsOnGene.append(line)
    
            ### chromosome numbers of current snp and gene are match
            else:
                ### current snp on current gene
                if int(np_UCSCGenomeDatabase[idx_gene, 1]) <= int_position and int_position <= int(np_UCSCGenomeDatabase[idx_gene, 2]):
                    list_snpsOnGene.append(line)
                ### snp position exceed this gene
                elif int_position > int(np_UCSCGenomeDatabase[idx_gene, 2]):
                    if len(list_snpsOnGene) != 0:
                        #### write gen file of current gene (output file name: geneSymbol_numOfSNPOnGene.gen)
                        #str_outputFileName = str(np_UCSCGenomeDatabase[idx_gene, 4]) + "_" + str(len(list_snpsOnGene)) + ".gen"
                        #with open(os.path.join(str_outputFilePath, str_outputFileName), "w") as file_outputFile:
                        #    for item in list_snpsOnGene:
                        #        file_outputFile.writelines(item)
                        str_outputFileName = str(np_UCSCGenomeDatabase[idx_gene, 4]) + "_" + str(len(list_snpsOnGene))
                        SplitMegaGene(list_snpsOnGene, int_window, int_step, str_outputFilePath, str_outputFileName)
                    list_snpsOnGene = []
                    while int_position > int(np_UCSCGenomeDatabase[idx_gene, 2]) and int_chromosome == int(np_UCSCGenomeDatabase[idx_gene, 0]):
                        ### jump to next gene
                        idx_gene = idx_gene + 1
                        ### if no next gene then break
                        if idx_gene == np_UCSCGenomeDatabase.shape[0]:
                            break
                        ### snp on next gene
                        if int(np_UCSCGenomeDatabase[idx_gene, 1]) <= int_position and int_position <= int(np_UCSCGenomeDatabase[idx_gene, 2]) and int_chromosome == int(np_UCSCGenomeDatabase[idx_gene, 0]):
                            list_snpsOnGene.append(line)

            ### if the index of gene out of the boundary of DB then break
            if idx_gene >= np_UCSCGenomeDatabase.shape[0]:
                break
    
    print("step3: Split by gene. DONE!")



# -*- coding: utf-8 -*-
"""
Created on Feb 2018

@author: Chester (Yu-Chuan Chang)
"""

""""""""""""""""""""""""""""""
# import libraries
""""""""""""""""""""""""""""""
import os
import warnings
warnings.filterwarnings('ignore')
# ignore all warnings
warnings.simplefilter("ignore")
os.environ["PYTHONWARNINGS"] = "ignore"

import os
import numpy as np
np.seterr(divide='ignore', invalid='ignore')

from sklearn import linear_model
from scipy.sparse import coo_matrix
from sklearn.utils import shuffle
from sklearn.model_selection import KFold
from sklearn.model_selection import GridSearchCV
from sklearn.externals import joblib

from genepi.step4_singleGeneEpistasis_Logistic import LogisticRegressionL1CV
from genepi.step4_singleGeneEpistasis_Lasso import LassoRegressionCV
from genepi.step5_crossGeneEpistasis_Logistic import LogisticRegressionL1
from genepi.step5_crossGeneEpistasis_Lasso import LassoRegression

""""""""""""""""""""""""""""""
# define functions 
""""""""""""""""""""""""""""""
def LoadDataForEnsemble(str_inputFileName_feature, str_inputFileName_phenotype):
    """

    Loading genetic features for ensembling with covariates

    Args:
        str_inputFilePath_feature (str): File path of input feature files from stage 2 - crossGeneEpistasis
        str_inputFileName_phenotype (str): File name of input phenotype data
        
    Returns:
        (tuple): tuple containing:

            - np_genotype (ndarray): 2D array containing genotype data with `int8` type
            - np_phenotype (ndarray): 2D array containing phenotype data with `float` type
    
    """

    ### get all selected snp ids
    list_genotype_rsid = []
    with open(str_inputFileName_feature, "r") as file_inputFile:
        ### grep the header
        list_rsids = file_inputFile.readline().strip().split(",")
        for rsid in list_rsids:
            list_genotype_rsid.append(rsid)
    np_genotype_rsid = np.array(list_genotype_rsid)
    
    ### count lines of input files
    int_num_genotype = len(np_genotype_rsid)
    int_num_phenotype = sum(1 for line in open(str_inputFileName_phenotype))
    
    ### get phenotype file
    list_phenotype = []
    with open(str_inputFileName_phenotype, 'r') as file_inputFile:
        for line in file_inputFile:
            list_phenotype.append(line.strip().split(","))
    np_phenotype = np.array(list_phenotype)
    del list_phenotype
    
    if np_phenotype.shape[1] < 2:
        print("step6: There is no other factors exist.")
        return None, None
    
    ### get genotype file
    np_genotype = np.empty([int_num_phenotype, int_num_genotype], dtype=np.float16)
    with open(str_inputFileName_feature, "r") as file_inputFile:
        ### skip header
        file_inputFile.readline()
        idx_phenotype = 0
        ### read feaure and write into np_genotype
        for line in file_inputFile:
            np_genotype[idx_phenotype, :len(list_rsids)] = np.array([float(x) for x in line.strip().split(",")], dtype='int')
            idx_phenotype = idx_phenotype + 1
    
    ### concatenate genotype and other factors
    np_genotype = np.concatenate((np_genotype, np_phenotype[:, :-1]), axis=1).astype(float)
    
    return np_genotype, np_phenotype

def ClassifierModelPersistence(np_X, np_y, str_outputFilePath = "", int_nJobs = 4):
    """

    Dumping ensemble classifier for model persistence

    Args:
        np_X (ndarray): 2D array containing genotype data with `int8` type
        np_y (ndarray): 2D array containing phenotype data with `float` type
        str_outputFilePath (str): File path of output file
        int_nJobs (int): The number of thread (default: 4)

    Returns:
        None
    
    """

    X = np_X
    y = np_y
    X_sparse = coo_matrix(X)
    X, X_sparse, y = shuffle(X, X_sparse, y, random_state=0)
    
    cost = [2**x for x in range(-8, 8)]
    parameters = [{'C':cost, 'penalty':['l1'], 'dual':[False], 'class_weight':['balanced']}]
    kf_estimator = KFold(n_splits=2)
    estimator_logistic = linear_model.LogisticRegression(max_iter=100, solver='liblinear')
    estimator_grid = GridSearchCV(estimator_logistic, parameters, scoring='f1', n_jobs=int_nJobs, cv=kf_estimator)
    estimator_grid.fit(X, y)
    
    joblib.dump(estimator_grid.best_estimator_, os.path.join(str_outputFilePath, "Classifier_Covariates.pkl"))

def RegressorModelPersistence(np_X, np_y, str_outputFilePath = "", int_nJobs = 4):
    """

    Dumping ensemble regressor for model persistence

    Args:
        np_X (ndarray): 2D array containing genotype data with `int8` type
        np_y (ndarray): 2D array containing phenotype data with `float` type
        str_outputFilePath (str): File path of output file
        int_nJobs (int): The number of thread (default: 4)

    Returns:
        None
    
    """
    
    X = np_X
    y = np_y
    X_sparse = coo_matrix(X)
    X, X_sparse, y = shuffle(X, X_sparse, y, random_state=0)
    
    alpha = np.logspace(-10, 10, 200)
    parameters = [{'alpha':alpha}]
    kf_estimator = KFold(n_splits=2)
    estimator_lasso = linear_model.Lasso(max_iter=1000)
    estimator_grid = GridSearchCV(estimator_lasso, parameters, scoring='neg_mean_squared_error', n_jobs=int_nJobs, cv=kf_estimator)
    estimator_grid.fit(X, y)
    
    joblib.dump(estimator_grid.best_estimator_, os.path.join(str_outputFilePath, "Regressor_Covariates.pkl"))

""""""""""""""""""""""""""""""
# main function
""""""""""""""""""""""""""""""
def EnsembleWithCovariatesClassifier(str_inputFileName_feature, str_inputFileName_phenotype, str_outputFilePath = "", int_kOfKFold = 2, int_nJobs = 4):
    """

    A workflow to ensemble genetic features with covariates for L1-regularized Logistic regression.

    Args:
        str_inputFilePath_feature (str): File path of input feature files from stage 2 - crossGeneEpistasis
        str_inputFileName_phenotype (str): File name of input phenotype data
        str_outputFilePath (str): File path of output file
        int_kOfKFold (int): The k for k-fold cross validation (default: 2)
        int_nJobs (int): The number of thread (default: 4)

    Returns:
        (tuple): tuple containing:

            - float_f1Score_train (float): The F1 score of the model for training set
            - float_f1Score_test (float): The F1 score of the model for testing set

        - Expected Success Response::

            "step6: Ensemble with covariates. DONE!"
    
    """
    
    ### set default output path
    if str_outputFilePath == "":
        str_outputFilePath = os.path.dirname(str_inputFileName_feature)
    
    #-------------------------
    # load data
    #-------------------------
    np_genotype, np_phenotype = LoadDataForEnsemble(str_inputFileName_feature, str_inputFileName_phenotype)
    if np_genotype is None and np_phenotype is None:
        return 0.0, 0.0
    
    #-------------------------
    # build model
    #-------------------------
    float_f1Score_test, np_weight = LogisticRegressionL1CV(np_genotype, np_phenotype[:, -1].astype(int), int_kOfKFold, int_nJobs)
    float_f1Score_train = LogisticRegressionL1(np_genotype, np_phenotype[:, -1].astype(int), int_nJobs)
    
    #-------------------------
    # dump persistent model
    #-------------------------
    ClassifierModelPersistence(np_genotype, np_phenotype[:, -1].astype(int), str_outputFilePath, int_nJobs)
    
    print("step6: Ensemble with covariates. DONE! (Training score:" + "{0:.2f}".format(float_f1Score_train) + "; " + str(int_kOfKFold) + "-fold Test Score:" + "{0:.2f}".format(float_f1Score_test) + ")")

    return float_f1Score_train, float_f1Score_test

def EnsembleWithCovariatesRegressor(str_inputFileName_feature, str_inputFileName_phenotype, str_outputFilePath = "", int_kOfKFold = 2, int_nJobs = 4):
    """

    A workflow to ensemble genetic features with covariates for L1-regularized Lasso regression.

    Args:
        str_inputFilePath_feature (str): File path of input feature files from stage 2 - crossGeneEpistasis
        str_inputFileName_phenotype (str): File name of input phenotype data
        str_outputFilePath (str): File path of output file
        int_kOfKFold (int): The k for k-fold cross validation (default: 2)
        int_nJobs (int): The number of thread (default: 4)

    Returns:
        (tuple): tuple containing:

            - float_AVG_S_P_train (float): The average of the Peason's and Spearman's correlation of the model for training set
            - float_AVG_S_P_test (float): The average of the Peason's and Spearman's correlation of the model for testing set

        - Expected Success Response::

            "step6: Ensemble with covariates. DONE!"
    
    """
    
    ### set default output path
    if str_outputFilePath == "":
        str_outputFilePath = os.path.dirname(str_inputFileName_feature)
    
    #-------------------------
    # load data
    #-------------------------
    np_genotype, np_phenotype = LoadDataForEnsemble(str_inputFileName_feature, str_inputFileName_phenotype)
    if np_genotype is None and np_phenotype is None:
        return 0.0, 0.0
    
    #-------------------------
    # build model
    #-------------------------
    float_AVG_S_P_test, np_weight = LassoRegressionCV(np_genotype, np_phenotype[:, -1].astype(float), int_kOfKFold, int_nJobs)
    float_AVG_S_P_train = LassoRegression(np_genotype, np_phenotype[:, -1].astype(float), int_nJobs)
    
    #-------------------------
    # dump persistent model
    #-------------------------
    RegressorModelPersistence(np_genotype, np_phenotype[:, -1].astype(int), str_outputFilePath, int_nJobs)
    
    print("step6: Ensemble with covariates. DONE! (Training score:" + "{0:.2f}".format(float_AVG_S_P_train) + "; " + str(int_kOfKFold) + "-fold Test Score:" + "{0:.2f}".format(float_AVG_S_P_test) + ")")
    
    return float_AVG_S_P_train, float_AVG_S_P_test


# -*- coding: utf-8 -*-
"""
Created on Apr 2019

@author: Chester (Yu-Chuan Chang)
"""

""""""""""""""""""""""""""""""
# import libraries
""""""""""""""""""""""""""""""
import argparse
import os
import numpy as np
import scipy.stats as sp
import pandas as pd
import matplotlib.pyplot as plt

""""""""""""""""""""""""""""""
# define functions 
""""""""""""""""""""""""""""""
def ArgumentsParser():
    ### define arguments
    str_description = ''
    'This script is a visualization tool for displaying the potential rare variants in vcf file'
    parser = argparse.ArgumentParser(prog='vcf2plot', description=str_description)
    
    ### define arguments for I/O
    parser.add_argument("-v", required=True, help="filename of the input vcf file")
    parser.add_argument("-p", required=True, help="filename of the input phenotype")    
    parser.add_argument("-o", required=False, help="output file path")
    
    ### define other optional arguments
    parser.add_argument("-m", required=False, default="r", choices=["c", "r"], help="choose phenotype type: c for classification; r for regression")
    parser.add_argument("-t", required=False, default=0.0, help="the threshold for categorize")
    
    return parser

def DrawRegressionPlot(df_variant, df_data, float_threshold, str_outputFilePath):
    ### draw scatter plot for real number phenotype
    plt.figure(figsize=(15, 40))
    plt.scatter(df_data["phenotype"], df_data["order"], color=df_data["color"], s=40, marker="o", alpha=0.5)
    plt.ylim(-1 , 100)
    plt.yticks(df_data["order"], df_data["variant_ID_x"])
    for idx_y in df_data["order"]:
        plt.axhline(y=idx_y, color="#CCCCCC", linewidth=1)
    plt.axvline(x=float_threshold, color="#CCCCCC", linewidth=1)
    plt.savefig(os.path.join(str_outputFilePath, "vcf2plot.jpg"), bbox_inches='tight', dpi=100)

def DrawClassificationPlot(df_variant, df_data, str_outputFilePath):
    ### count the number of mutations in each variant
    df_data = df_data.groupby(["order", "variant_ID_x", "phenotype"])["allele_type"].sum().unstack(fill_value=0).reset_index()
    ### draw barplot plot for dualistic phenotype
    float_barwidth = 0.35
    float_opacity = 0.8
    fig, ax = plt.subplots(1, 1, figsize=(15, 40))
    plt.barh(df_data["order"] + float_barwidth * 0.5, df_data.iloc[:,-1], float_barwidth, alpha=float_opacity, color="#CC2900", label='Case')
    plt.barh(df_data["order"] + float_barwidth * 1.5, df_data.iloc[:,-2], float_barwidth, alpha=float_opacity, color="#29A329", label='Control')
    plt.ylim(-1 , 100)
    plt.yticks(df_data["order"] + float_barwidth, df_data["variant_ID_x"])
    for idx_y in df_data["order"]:
        plt.axhline(y=idx_y + float_barwidth * 1, color="#CCCCCC", linewidth=1)
    plt.legend() 
    plt.savefig(os.path.join(str_outputFilePath, "vcf2plot.jpg"), bbox_inches='tight', dpi=100)

""""""""""""""""""""""""""""""
# main function
""""""""""""""""""""""""""""""
def main(args=None):
    ### obtain arguments from argument parser
    args = ArgumentsParser().parse_args(args)

    ### get arguments for I/O
    str_inputFileName_vcf = args.v
    str_inputFileName_phenotype = args.p
    str_outputFilePath = ""
    if args.o is not None:
        str_outputFilePath = args.o
    else:
        str_outputFilePath = os.path.dirname(str_inputFileName_vcf)
    float_threshold = float(args.t)    

    ### get phenotype of each sample
    dict_pheno = {}
    with open(str_inputFileName_phenotype, "r") as file_inputFile:
        ### skip two headers
        file_inputFile.readline()
        for line in file_inputFile:
            list_line = line.strip().split("\t")
            list_line[1] = "{:010d}".format(int(list_line[1]))
            dict_pheno[list_line[1]] = list_line[-1]
    
    ### scan vcf file
    # [idx, variant ID, odds ratio, p-value, mutation in control, mutation in case]
    np_variant = []
    # [idx, variant ID, sample ID, allele type, phenotype]
    np_data = []
    idx_variant = 0
    with open(str_inputFileName_vcf, "r") as file_inputFile:
        for line in file_inputFile:
            ### skip headers
            if line[0] == "#":
                list_sample_id = line.strip().split("\t")
            else:
                ### create contigency table for fisher-exact test
                np_contingency = np.array([[0, 0], [0, 0]])
                list_line = line.strip().split("\t")
                ### skip low quality variants
                if list_line[6] != "LowQual":
                    ### grep genotype of each sample
                    for idx_sample in range(len(list_line[9:])):
                        str_sample = list_sample_id[idx_sample + 9].split("_")[1]
                        if str_sample in dict_pheno.keys():
                            list_item = list_line[idx_sample + 9].split(":")
                            ### if the sample be classified as class I
                            if float(dict_pheno[str_sample]) > float_threshold:
                                if list_item[0] == ".":
                                    np_contingency[0, 1] += 2
                                elif list_item[0] == "0/1":
                                    np_contingency[0, 1] += 1
                                    np_contingency[1, 1] += 1
                                    np_data.append([idx_variant, str(list_line[0] + ":" + list_line[1]), str_sample, 1, dict_pheno[str_sample]])                        
                                elif list_item[0] == "1/1":
                                    np_contingency[1, 1] += 2
                                    np_data.append([idx_variant, str(list_line[0] + ":" + list_line[1]), str_sample, 2, dict_pheno[str_sample]])
                                else:
                                    np_contingency[0, 1] += 2
                            ### the sample is belong class II
                            else:
                                if list_item[0] == ".":
                                    np_contingency[0, 0] += 2
                                elif list_item[0] == "0/1":
                                    np_contingency[0, 0] += 1
                                    np_contingency[1, 0] += 1
                                    np_data.append([idx_variant, str(list_line[0] + ":" + list_line[1]), str_sample, 1, dict_pheno[str_sample]])                        
                                elif list_item[0] == "1/1":
                                    np_contingency[1, 0] += 2
                                    np_data.append([idx_variant, str(list_line[0] + ":" + list_line[1]), str_sample, 2, dict_pheno[str_sample]])
                                else:
                                    np_contingency[0, 0] += 2
                    ### execute fisher-exact test
                    oddsratio, pvalue = sp.fisher_exact(np_contingency)
                    np_variant.append([idx_variant, str(list_line[0] + ":" + list_line[1]), oddsratio, pvalue, np_contingency[1, 0], np_contingency[1, 1]])
                    idx_variant += 1
    
    ### prepare variant's info for plot
    df_variant = pd.DataFrame(np_variant)
    df_variant.columns = ["idx", "variant_ID", "odds_ratio", "p-value", "mutation_in_control", "mutation_in_case"]
    df_variant = df_variant.assign(mutation_sum = df_variant["mutation_in_control"] + df_variant["mutation_in_case"])
    df_variant = df_variant.sort_values(by=["odds_ratio", "mutation_sum"])
    df_variant = df_variant.dropna(subset=["odds_ratio"])
    ### obtain top 100 significant variants
    if df_variant.shape[0] > 100:
        df_temp = df_variant[df_variant["odds_ratio"] < 1].sort_values(by=["odds_ratio", "mutation_sum"], ascending=[True, False])
        df_temp = pd.concat([df_temp, df_variant[df_variant["odds_ratio"] > 1]])
        df_variant = df_temp
        df_temp = df_variant[:50]
        df_temp = pd.concat([df_temp, df_variant[-50:]])
        df_variant = df_temp
    ### set order of variants for plot
    df_variant = df_variant.assign(order = range(df_variant.shape[0]))
    df_variant["label"]= df_variant["odds_ratio"].map('{:,.4f}'.format) + "-" + df_variant["mutation_in_control"].map(str) + ":" + df_variant["mutation_in_case"].map(str)
    
    ### prepare mutations's info for plot
    df_data = pd.DataFrame(np_data)
    df_data.columns = ["idx", "variant_ID", "sample_ID", "allele_type", "phenotype"]
    df_data["phenotype"] = df_data["phenotype"].astype("float")
    df_data = pd.merge(df_data, df_variant, how="inner", on="idx")
    ### set color of mutations for plot
    df_data["color"] = np.where(df_data["phenotype"] > float_threshold, "#CC2900", "#29A329")
    
    ### draw plot
    if args.m=="c":
        DrawClassificationPlot(df_variant, df_data, str_outputFilePath)
    else:
        DrawRegressionPlot(df_variant, df_data, float_threshold, str_outputFilePath)

if __name__ == "__main__":
    main()


# -*- coding: utf-8 -*-
"""
Created on Feb 2018

@author: Chester (Yu-Chuan Chang)
"""

import os
import genepi


def main():
    # give file names of "Genotype" and "Phenotype" here !
    str_inputFileName_genotype = os.path.dirname(os.path.abspath(__file__)) + "/sample.gen"
    str_inputFileName_phenotype = os.path.dirname(os.path.abspath(__file__)) + "/sample.csv"
    
    ### step1_downloadUCSCDB
    genepi.DownloadUCSCDB(str_hgbuild="hg19")
    
    ### step2_estimateLD
    genepi.EstimateLDBlock(str_inputFileName_genotype, float_threshold_DPrime=0.9, float_threshold_RSquare=0.9)
    
    ### step3_splitByGene
    genepi.SplitByGene(str_inputFileName_genotype.replace(".gen", "_LDReduced.gen"))
    
    ### step4_singleGeneEpistasis_Logistic (for case/control trial)
    genepi.BatchSingleGeneEpistasisLogistic(os.path.dirname(str_inputFileName_genotype) + "/snpSubsets/", str_inputFileName_phenotype, int_nJobs=1)
    ### step4_singleGeneEpistasis_Lasso (for quantitative trial)
    #genepi.BatchSingleGeneEpistasisLasso(os.path.dirname(str_inputFileName_genotype) + "/snpSubsets/", str_inputFileName_phenotype, int_nJobs=1)
    
    ### step5_crossGeneEpistasis_Logistic (for case/control trial)
    genepi.CrossGeneEpistasisLogistic(os.path.dirname(str_inputFileName_genotype) + "/singleGeneResult/", str_inputFileName_phenotype, int_nJobs=1)
    ### step5_crossGeneEpistasis_Lasso (for quantitative trial)
    #genepi.CrossGeneEpistasisLasso(os.path.dirname(str_inputFileName_genotype) + "/singleGeneResult/", str_inputFileName_phenotype, int_nJobs=1)
    
    ### step6_ensembleWithCovariates (for case/control trial)
    genepi.EnsembleWithCovariatesClassifier(os.path.dirname(str_inputFileName_genotype) + "/crossGeneResult/Feature.csv", str_inputFileName_phenotype, int_nJobs=1)
    ### step6_ensembleWithCovariates (for quantitative trial)
    #genepi.EnsembleWithCovariatesRegressor(os.path.dirname(str_inputFileName_genotype) + "/crossGeneResult/Feature.csv", str_inputFileName_phenotype, int_nJobs=1)
    
if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
"""
Created on Feb 2018

@author: Chester (Yu-Chuan Chang)
"""

""""""""""""""""""""""""""""""
# import libraries
""""""""""""""""""""""""""""""
from setuptools import setup, find_packages

with open('README.md') as f:
    long_description = f.read()

setup(
    name = 'genepi',
    version = '2.0.4',
    description = 'A package for detecting epsitasis by machine learning',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/Chester75321/GenEpi',
    author = 'Chester (Yu-Chuan Chang)',
    author_email = 'chester75321@gmail.com',
    license = 'MIT',
    project_urls={
        "Source Code": "https://github.com/Chester75321/GenEpi",
        "Documentation": "https://genepi.readthedocs.io/en/latest/",
        "Bug Tracker": "https://github.com/Chester75321/GenEpi/issues",
    },
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
    ],
    keywords = ['epistasis', 'SNP-SNP interactions', 'GWAS'],
    packages = find_packages(),
    install_requires=[
        'pymysql>=0.8.0',
        'numpy>=1.13.0',
        'scipy>=0.19.0',
        'psutil>=4.3.0',
        'scikit-learn>=0.21.2',
        'matplotlib>=3.1.1',
        'pandas>=0.25.2',
        'seaborn>=0.9.0',
    ],
    python_requires='>=3',
    include_package_data = True,
    zip_safe = False,
    entry_points = {
        'console_scripts': ['GenEpi=genepi.GenEpi:main'],
    }
)



# -*- coding: utf-8 -*-
"""
Created on Feb 2018

@author: Chester (Yu-Chuan Chang)
"""

""""""""""""""""""""""""""""""
# import libraries
""""""""""""""""""""""""""""""
import os
import warnings
warnings.filterwarnings('ignore')
# ignore all warnings
warnings.simplefilter("ignore")
os.environ["PYTHONWARNINGS"] = "ignore"

import os
import numpy as np
import scipy as sp
np.seterr(divide='ignore', invalid='ignore')
from sklearn.feature_selection import chi2
from sklearn import linear_model
from scipy.sparse import coo_matrix
from sklearn.utils import shuffle
from sklearn.model_selection import KFold
from sklearn.model_selection import GridSearchCV
from sklearn.externals import joblib
import sklearn.metrics as skMetric
import scipy.stats as stats

import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns

from genepi.step4_singleGeneEpistasis_Logistic import RandomizedLogisticRegression
from genepi.step4_singleGeneEpistasis_Logistic import LogisticRegressionL1CV
from genepi.step4_singleGeneEpistasis_Logistic import FeatureEncoderLogistic

""""""""""""""""""""""""""""""
# define functions 
""""""""""""""""""""""""""""""
def LogisticRegressionL1(np_X, np_y, int_nJobs = 4):
    """

    Implementation of the L1-regularized Logistic regression with k-fold cross validation.

    Args:
        np_X (ndarray): 2D array containing genotype data with `int8` type
        np_y (ndarray): 2D array containing phenotype data with `float` type
        int_nJobs (int): The number of thread (default: 4)

    Returns:
        (float): float_f1Score
        
            The F1 score of the model
    
    """

    X = np_X
    y = np_y
    
    list_target = []
    list_predict = []
    
    cost = [2**x for x in range(-8, 8)]
    parameters = [{'C':cost, 'penalty':['l1'], 'dual':[False], 'class_weight':['balanced']}]
    kf_estimator = KFold(n_splits=2)
    estimator_logistic = linear_model.LogisticRegression(max_iter=100, solver='liblinear')
    estimator_grid = GridSearchCV(estimator_logistic, parameters, scoring='f1', n_jobs=int_nJobs, cv=kf_estimator)
    estimator_grid.fit(X, y)
    list_label = estimator_grid.best_estimator_.predict(X)
    for idx_y, idx_label in zip(list(y), list_label):
        list_target.append(float(idx_y))
        list_predict.append(idx_label)
    float_f1Score = skMetric.f1_score(list_target, list_predict)
    
    return float_f1Score

def GenerateContingencyTable(np_genotype, np_phenotype):
    """

    Generating the contingency table for chi-square test.

    Args:
        np_X (ndarray): 2D array containing genotype data with `int8` type
        np_y (ndarray): 2D array containing phenotype data with `float` type

    Returns:
        (ndarray): np_contingency 
        
            2D array containing the contingency table with `int` type
    
    """

    np_contingency = np.array([[0, 0], [0, 0]])
    for idx_subject in range(0, np_genotype.shape[0]):
        np_contingency[int(np_genotype[idx_subject]), int(np_phenotype[idx_subject])] = np_contingency[int(np_genotype[idx_subject]), int(np_phenotype[idx_subject])] + 1
    np_contingency = np.rot90(np_contingency)
    np_contingency = np.rot90(np_contingency)
    
    return np_contingency

def ClassifierModelPersistence(np_X, np_y, str_outputFilePath = "", int_nJobs = 4):
    """

    Dumping classifier for model persistence

    Args:
        np_X (ndarray): 2D array containing genotype data with `int8` type
        np_y (ndarray): 2D array containing phenotype data with `float` type
        str_outputFilePath (str): File path of output file
        int_nJobs (int): The number of thread (default: 4)

    Returns:
        None
    
    """

    X = np_X
    y = np_y
    X_sparse = coo_matrix(X)
    X, X_sparse, y = shuffle(X, X_sparse, y, random_state=0)
    
    cost = [2**x for x in range(-8, 8)]
    parameters = [{'C':cost, 'penalty':['l1'], 'dual':[False], 'class_weight':['balanced']}]
    kf_estimator = KFold(n_splits=2)
    estimator_logistic = linear_model.LogisticRegression(max_iter=100, solver='liblinear')
    estimator_grid = GridSearchCV(estimator_logistic, parameters, scoring='f1', n_jobs=int_nJobs, cv=kf_estimator)
    estimator_grid.fit(X, y)
    
    joblib.dump(estimator_grid.best_estimator_, os.path.join(str_outputFilePath, "Classifier.pkl"))

def fsigmoid(x, a, b):
    float_return = 1.0/(1.0+np.exp(-a*(x-b)))
    return float_return

def PlotPolygenicScore(np_X, np_y, int_kOfKFold = 2, int_nJobs = 1, str_outputFilePath=""):
    """

    Plot figure for polygenic score, including group distribution and prevalence to PGS

    Args:
        np_X (ndarray): 2D array containing genotype data with `int8` type
        np_y (ndarray): 2D array containing phenotype data with `float` type
        int_kOfKFold (int): The k for k-fold cross validation (default: 2)
        int_nJobs (int): The number of thread (default: 1)

    Returns:
        None
    
    """

    X = np_X
    y = np_y
    X_sparse = coo_matrix(X)
    X, X_sparse, y = shuffle(X, X_sparse, y, random_state=0)
    kf = KFold(n_splits=int_kOfKFold)

    list_target = []
    list_predict = []
    list_proba = []
    for idxTr, idxTe in kf.split(X):
        cost = [2**x for x in range(-8, 8)]
        parameters = [{'C':cost, 'penalty':['l1'], 'dual':[False], 'class_weight':['balanced']}]
        kf_estimator = KFold(n_splits=2)
        estimator_logistic = linear_model.LogisticRegression(max_iter=100, solver='liblinear')
        estimator_grid = GridSearchCV(estimator_logistic, parameters, scoring='f1', n_jobs=1, cv=kf_estimator)
        estimator_grid.fit(X[idxTr], y[idxTr])
        list_predict_this = estimator_grid.best_estimator_.predict(X[idxTe])
        list_proba_this = estimator_grid.best_estimator_.predict_proba(X[idxTe])
        for idx_y, idx_predict, idx_proba in zip(list(y[idxTe]), list_predict_this, list_proba_this):
            list_target.append(float(idx_y))
            list_predict.append(idx_predict)
            list_proba.append(idx_proba)
        
    float_f1Score = skMetric.f1_score(list_target, list_predict)

    #-------------------------
    # group distribution
    #-------------------------
    pd_pgs = pd.concat([pd.DataFrame(list_target), pd.DataFrame(list_predict), pd.DataFrame(np.array(list_proba)[:,1])], axis=1)
    pd_pgs.columns = ['target', 'predict', 'proba']

    pd_case = pd_pgs[pd_pgs.target == 1.0]
    sns.distplot(pd_case['proba'],  kde=True, bins=25, label='Case', color="#c9474b")
    pd_control = pd_pgs[pd_pgs.target == 0.0]
    sns.distplot(pd_control['proba'],  kde=True, bins=25, label='Control', color="#4e7e91")

    # plot formatting
    str_method = "GenEpi"
    plt.legend(prop={'size': 12})
    plt.title(str_method + ' on ADNI: ' + "%.4f" % float_f1Score + ' (F1 Score)')
    plt.xlim(0, 1)
    plt.ylim(0, 15)
    plt.xlabel('Polygenic Score')
    plt.ylabel('Frequency')
    plt.savefig(os.path.join(str_outputFilePath, str_method + "_PGS.png"), dpi=300)
    plt.close('all')

    #-------------------------
    # prevalence to PGS
    #-------------------------
    int_step = 5
    np_percentile = np.percentile(pd_pgs['proba'], q=list(range(0, 100 + int_step, int_step)))
    pd_pgs['percentile'] = np.searchsorted(np_percentile, pd_pgs['proba'], side='left') * int_step

    pd_prevalence_obs = pd_pgs.groupby('percentile').sum()[['target']]
    pd_prevalence_obs_count = pd_pgs.groupby('percentile').count()[['target']]
    pd_prevalence_obs = pd_prevalence_obs/pd_prevalence_obs_count
    pd_prevalence_obs.columns = ['obs']
    pd_prevalence_pre = pd_pgs.groupby('percentile').mean()[['proba']]
    pd_prevalence_pre.columns = ['pre']

    pd_rr_ingroup_case = pd_pgs.groupby('percentile').sum()[['target']]
    pd_rr_ingroup_sum = pd_pgs.groupby('percentile').count()[['target']]
    pd_rr = (pd_rr_ingroup_case / pd_rr_ingroup_sum) / (pd_pgs.sum()[['target']] / pd_pgs.count()[['target']])
    pd_rr.columns = ['Relative Risk']

    sns.scatterplot(x=pd_prevalence_obs.index, y=pd_prevalence_obs['obs'], hue=pd_rr['Relative Risk'], palette=sns.cubehelix_palette(8, start=.5, rot=-.75, as_cmap=True))
    popt, pcov = sp.optimize.curve_fit(fsigmoid, pd_prevalence_pre.index, pd_prevalence_pre['pre'], method='dogbox', bounds=([0., 0.],[1., 100.]))
    sns.lineplot(x=pd_prevalence_pre.index, y=fsigmoid(pd_prevalence_pre.index, *popt), color="black")

    plt.legend(prop={'size': 12}, loc='upper left')
    plt.title(str_method + ' on ADNI: ' + "%.4f" % float_f1Score + ' (F1 Score)')
    plt.xlim(0, 100)
    plt.ylim(0, 1)
    plt.xlabel('Polygenic Score Percentile')
    plt.ylabel('Prevalence of Percentile Group')
    plt.savefig(os.path.join(str_outputFilePath, str_method + "_Prevalence.png"), dpi=300)
    plt.close('all')

""""""""""""""""""""""""""""""
# main function
""""""""""""""""""""""""""""""
def CrossGeneEpistasisLogistic(str_inputFilePath_feature, str_inputFileName_phenotype, str_inputFileName_score = "", str_outputFilePath = "", int_kOfKFold = 2, int_nJobs = 4):   
    """

    A workflow to model a cross gene epistasis containing two-element combinatorial encoding, stability selection, filtering low quality varaint and  L1-regularized Logistic regression with k-fold cross validation.

    Args:
        str_inputFilePath_feature (str): File path of input feature files from stage 1 - singleGeneEpistasis
        str_inputFileName_phenotype (str): File name of input phenotype data
        str_inputFileName_score (str): File name of input score file from stage 1 - singleGeneEpistasis
        str_outputFilePath (str): File path of output file
        int_kOfKFold (int): The k for k-fold cross validation (default: 2)
        int_nJobs (int): The number of thread (default: 4)

    Returns:
        (tuple): tuple containing:

            - float_f1Score_train (float): The F1 score of the model for training set
            - float_f1Score_test (float): The F1 score of the model for testing set

        - Expected Success Response::

            "step5: Detect cross gene epistasis. DONE!"
    
    """
    
    ### set default output path
    if str_outputFilePath == "":
        str_outputFilePath = os.path.abspath(os.path.join(str_inputFilePath_feature, os.pardir)) + "/crossGeneResult/"
    ### if output folder doesn't exist then create it
    if not os.path.exists(str_outputFilePath):
        os.makedirs(str_outputFilePath)
    
    ### set default score file name
    if str_inputFileName_score == "":
        for str_fileName in os.listdir(str_inputFilePath_feature):
            if str_fileName.startswith("All_Logistic"):
                str_inputFileName_score = os.path.join(str_inputFilePath_feature, str_fileName)

    #-------------------------
    # load data
    #-------------------------
    ### scan score file and exclude useless genes
    dict_score = {}
    with open(str_inputFileName_score, "r") as file_inputFile:
        file_inputFile.readline()
        for line in file_inputFile:
            list_thisScore = line.strip().split(",")
            if list_thisScore[1] == "MemErr" or float(list_thisScore[1]) == 0.0:
                pass
            else:
                dict_score[list_thisScore[0]] = float(list_thisScore[1])
    
    ### get all the file names of feature file
    list_featureFileName = []
    for str_fileName in os.listdir(str_inputFilePath_feature):
        if "Feature.csv" in str_fileName:
            list_featureFileName.append(str_fileName)
    
    ### get all selected snp ids
    list_genotype_rsid = []
    for item in list_featureFileName:
        with open(os.path.join(str_inputFilePath_feature, item), "r") as file_inputFile:
            ### grep the header
            list_rsids = file_inputFile.readline().strip().split(",")
            for rsid in list_rsids:
                list_genotype_rsid.append(rsid)
    np_genotype_rsid = np.array(list_genotype_rsid)
    
    ### count lines of input files
    int_num_genotype = len(np_genotype_rsid)
    int_num_phenotype = sum(1 for line in open(str_inputFileName_phenotype))
    
    ### get phenotype file
    list_phenotype = []
    with open(str_inputFileName_phenotype, 'r') as file_inputFile:
        for line in file_inputFile:
            list_phenotype.append(line.strip().split(","))
    np_phenotype = np.array(list_phenotype, dtype=np.float)
    del list_phenotype
    
    ### get genotype file
    ### declare a dictionary for mapping snp and gene
    dict_geneMap ={}
    idx_genotype_rsid = 0
    np_genotype = np.empty([int_num_phenotype, int_num_genotype], dtype='int8')
    for item in list_featureFileName:
        with open(os.path.join(str_inputFilePath_feature, item), "r") as file_inputFile:
            ### grep feature from header of feature file
            list_rsids = file_inputFile.readline().strip().split(",")
            for rsid in list_rsids:
                ### key: rsIDs of a feature; value: gene symbol
                dict_geneMap[rsid] = item.split("_")[0]
            idx_phenotype = 0
            ### read feaure and write into np_genotype
            for line in file_inputFile:
                np_genotype[idx_phenotype, idx_genotype_rsid:idx_genotype_rsid + len(list_rsids)] = np.array([float(x) for x in line.strip().split(",")], dtype='int')
                idx_phenotype = idx_phenotype + 1
            idx_genotype_rsid = idx_genotype_rsid + len(list_rsids)
    
    #-------------------------
    # preprocess data
    #-------------------------
    ### chi-square test selection
    np_chi2 = -np.log10(chi2(np_genotype.astype(int), np_phenotype[:, -1].astype(int))[1])
    np_selectedIdx = np.array([x > 5 for x in np_chi2])
    np_genotype = np_genotype[:, np_selectedIdx]
    np_genotype_rsid = np_genotype_rsid[np_selectedIdx]
    if np_genotype_rsid.shape[0] == 0:
        return 0.0

    ### select degree 1 feature
    np_genotype_rsid_degree = np.array([str(x).count('*') + 1 for x in np_genotype_rsid])
    np_selectedIdx = np.array([x == 1 for x in np_genotype_rsid_degree])
    np_genotype_degree1 = np_genotype[:, np_selectedIdx]
    np_genotype_degree1_rsid = np_genotype_rsid[np_selectedIdx]
    
    ### remove redundant polynomial features
    np_genotype_degree1, np_selectedIdx = np.unique(np_genotype_degree1, axis=1, return_index=True)
    np_genotype_degree1_rsid = np_genotype_degree1_rsid[np_selectedIdx]
    
    ### generate cross gene interations
    np_genotype_crossGene_rsid, np_genotype_crossGene = FeatureEncoderLogistic(np_genotype_degree1_rsid, np_genotype_degree1, np_phenotype, 1)
    
    ### remove degree 1 feature from dataset
    np_selectedIdx = np.array([x != 1 for x in np_genotype_rsid_degree])
    np_genotype = np_genotype[:, np_selectedIdx]
    np_genotype_rsid = np_genotype_rsid[np_selectedIdx]
    
    ### concatenate cross gene interations
    if np_genotype_degree1.shape[1] > 0:
        np_genotype = np.concatenate((np_genotype, np_genotype_crossGene), axis=1)
        np_genotype_rsid = np.concatenate((np_genotype_rsid, np_genotype_crossGene_rsid))
    
    #-------------------------
    # select feature
    #-------------------------    
    ### random logistic feature selection
    np_randWeight = np.array(RandomizedLogisticRegression(np_genotype, np_phenotype[:, -1].astype(int)))    
    np_selectedIdx = np.array([x >= 0.25 for x in np_randWeight])
    np_randWeight = np_randWeight[np_selectedIdx]
    np_genotype = np_genotype[:, np_selectedIdx]
    np_genotype_rsid = np_genotype_rsid[np_selectedIdx]
    if np_genotype_rsid.shape[0] == 0:
        return 0.0
    
    #-------------------------
    # build model
    #-------------------------
    float_f1Score_test, np_weight = LogisticRegressionL1CV(np_genotype, np_phenotype[:, -1].astype(int), int_kOfKFold, int_nJobs)
    float_f1Score_train = LogisticRegressionL1(np_genotype, np_phenotype[:, -1].astype(int), int_nJobs)
    
    ### filter out zero-weight features
    np_selectedIdx = np.array([x != 0.0 for x in np_weight])
    np_weight = np_weight[np_selectedIdx]
    np_genotype = np_genotype[:, np_selectedIdx]
    np_genotype_rsid = np_genotype_rsid[np_selectedIdx]
    if np_genotype_rsid.shape[0] == 0:
        return 0.0
    
    #-------------------------
    # analyze result
    #-------------------------
    ### calculate chi-square p-value
    np_chi2 = -np.log10(chi2(np_genotype.astype(int), np_phenotype[:, -1].astype(int))[1])
    list_oddsRatio = []
    for idx_feature in range(0, np_genotype.shape[1]):
        np_contingency = GenerateContingencyTable(np_genotype[:, idx_feature], np_phenotype[:, -1])
        oddsratio, pvalue = stats.fisher_exact(np_contingency)
        list_oddsRatio.append(oddsratio)
        
    ### calculate genotype frequency
    np_genotypeFreq = np.sum(np_genotype, axis=0).astype(float) / np_genotype.shape[0]
    
    #-------------------------
    # output results
    #-------------------------
    ### output statistics of features
    with open(os.path.join(str_outputFilePath, "Result.csv"), "w") as file_outputFile:
        file_outputFile.writelines("rsid,weight,chi-square_log_p-value,odds_ratio,genotype_frequency,geneSymbol,singleGeneScore" + "\n")
        for idx_feature in range(0, np_genotype_rsid.shape[0]):
            ### if this feature is single gene epistasis
            if np_genotype_rsid[idx_feature,] in dict_geneMap.keys():
                str_thisOutput = str(np_genotype_rsid[idx_feature,]) + "," + str(np_weight[idx_feature,]) + "," + str(np_chi2[idx_feature,]) + "," + str(list_oddsRatio[idx_feature]) + "," + str(np_genotypeFreq[idx_feature]) + "," + str(dict_geneMap[np_genotype_rsid[idx_feature,]]).split("@")[0] + "," + str(dict_score[dict_geneMap[np_genotype_rsid[idx_feature,]]]) + "\n"
                file_outputFile.writelines(str_thisOutput)
            ### else this feature is cross gene epistasis
            else:
                str_thisOutput = str(np_genotype_rsid[idx_feature,]) + "," + str(np_weight[idx_feature,]) + "," + str(np_chi2[idx_feature,]) + "," + str(list_oddsRatio[idx_feature]) + "," + str(np_genotypeFreq[idx_feature]) + "," + str(dict_geneMap[np_genotype_rsid[idx_feature,].split("*")[0]]).split("@")[0] + "*" + str(dict_geneMap[np_genotype_rsid[idx_feature,].split("*")[1]]).split("@")[0] + ", " + "\n"
                file_outputFile.writelines(str_thisOutput)            
 
    ### output feature
    with open(os.path.join(str_outputFilePath, "Feature.csv"), "w") as file_outputFile:
        file_outputFile.writelines(",".join(np_genotype_rsid) + "\n")
        for idx_subject in range(0, np_genotype.shape[0]):
            file_outputFile.writelines(",".join(np_genotype[idx_subject, :].astype(str)) + "\n")

    ### output figures
    PlotPolygenicScore(np_genotype, np_phenotype[:, -1].astype(int), int_nJobs, str_outputFilePath)

    #-------------------------
    # dump persistent model
    #-------------------------
    ClassifierModelPersistence(np_genotype, np_phenotype[:, -1].astype(int), str_outputFilePath, int_nJobs)
    
    print("step5: Detect cross gene epistasis. DONE! (Training score:" + "{0:.2f}".format(float_f1Score_train) + "; " + str(int_kOfKFold) + "-fold Test Score:" + "{0:.2f}".format(float_f1Score_test) + ")")
    
    return float_f1Score_train, float_f1Score_test



# -*- coding: utf-8 -*-
"""
Created on Apr 2019

@author: Chester (Yu-Chuan Chang)
"""

""""""""""""""""""""""""""""""
# import libraries
""""""""""""""""""""""""""""""
import argparse
import time
import os
import sys
import multiprocessing as mp
from . import *

""""""""""""""""""""""""""""""
# define functions 
""""""""""""""""""""""""""""""
def ArgumentsParser():
    """

    To obtain and parse the arguments from user.

    Args:
        None

    Returns:
        argparse.ArgumentParser
    
    """

    ### define arguments
    str_description = ''
    'GenEpi is a package to uncover epistasis associated with phenotypes by a machine learning approach, '
    'developed by Yu-Chuan Chang at c4Lab of National Taiwan University. '
    'Github: https://github.com/Chester75321/GenEpi '
    'Reference: GenEpi: Gene-based Epistasis Discovery Using Machine Learning '
    '(https://www.biorxiv.org/content/10.1101/421719v5) '
    parser = argparse.ArgumentParser(prog='GenEpi', description=str_description)
    
    ### define arguments for I/O
    parser.add_argument("-g", required=True, help="filename of the input .gen file")
    parser.add_argument("-p", required=True, help="filename of the input phenotype")
    parser.add_argument("-s", required=False, help="self-defined genome regions")
    parser.add_argument("-o", required=False, help="output file path")
    
    ### define arguments for modeling
    parser.add_argument("-m", required=False, default="c", choices=["c", "r"], help="choose model type: c for classification; r for regression")
    parser.add_argument("-k", required=False, default=2, help="k of k-fold cross validation")
    parser.add_argument("-t", required=False, default=mp.cpu_count(), help="number of threads")
    
    ### define arguments for step1_downloadUCSCDB
    parser_group_1 = parser.add_argument_group("update UCSC database")
    parser_group_1.add_argument('--updatedb', action='store_true', default=False, help="enable this function")
    parser_group_1.add_argument("-b", required=False, default="hg19", choices=["hg19", "hg38"], help="human genome build")
    
    ### define arguments for step2_estimateLD
    parser_group_2 = parser.add_argument_group("compress data by LD block")
    parser_group_2.add_argument('--compressld', action='store_true', default=False, help="enable this function")
    parser_group_2.add_argument("-d", required = False, default=0.9, type=float, help="threshold for compression: D prime")
    parser_group_2.add_argument("-r", required = False, default=0.9, type=float, help="threshold for compression: R square")
 
    return parser

def InputChecking(str_inputFileName_genotype, str_inputFileName_phenotype):
    """

    To check the numbers of sample are consistent in genotype and phenotype data.

    Args:
        str_inputFileName_genotype (str): File name of input genotype data
        str_inputFileName_phenotype (str): File name of input phenotype data

    Returns:
        (tuple): tuple containing:

            - int_num_genotype (int): The sample number of genotype data
            - int_num_phenotype (int): The sample number of phenotype data 
    
    """

    ### check file name exist
    if str_inputFileName_genotype is None:
        sys.exit("There is no input genotype file.")
    if str_inputFileName_phenotype is None:
        sys.exit("There is no input phenotype file.")
    
    ### count lines of input files
    int_num_genotype = sum(1 for line in open(str_inputFileName_genotype))
    int_num_phenotype = sum(1 for line in open(str_inputFileName_phenotype))
    
    ### count sample num. in genotype file
    with open(str_inputFileName_genotype, 'r') as file_inputFile:
        list_line = file_inputFile.readline().strip().split(" ")
        int_num_genotype_sample = (len(list_line) - 5) / 3
        if int_num_genotype_sample != int_num_phenotype:
            sys.exit("The number of samples in genotype file does not match the number of samples in phenotype file.")
    
    return int_num_genotype, int_num_phenotype

""""""""""""""""""""""""""""""
# main function
""""""""""""""""""""""""""""""
def main(args=None):
    """

    Main function for obtaining user arguments, controling workflow and recording log file.

    Args:
        None

    Returns:
        None
    
    """
    ### obtain arguments from argument parser
    args = ArgumentsParser().parse_args(args)
    
    ### get arguments for I/O
    str_inputFileName_genotype = args.g
    str_inputFileName_phenotype = args.p
    str_inputFileName_regions = ""
    if args.s is not None:
        str_inputFileName_regions = args.s
    else:
        str_inputFileName_regions = "None"
    str_outputFilePath = ""
    if args.o is not None:
        str_outputFilePath = args.o
    else:
        str_outputFilePath = os.path.dirname(str_inputFileName_genotype)
    int_thread = mp.cpu_count()
    if int(args.t) is not None:
        if int(args.t) < mp.cpu_count():
            int_thread = int(args.t)
        
    if str_inputFileName_genotype == "example" and str_inputFileName_phenotype == "example":
        str_command = "cp " + os.path.join(os.path.dirname(__file__), "example", "sample.csv") + " " + str_outputFilePath
        os.system(str_command)
        str_command = "cp " + os.path.join(os.path.dirname(__file__), "example", "sample.gen") + " " + str_outputFilePath
        os.system(str_command)
        str_command = "GenEpi -g " + os.path.join(str_outputFilePath, "sample.gen") + " "
        str_command += "-p " + os.path.join(str_outputFilePath, "sample.csv") + " "
        str_command += "-o " + str_outputFilePath + " "
        str_command += "--updatedb --compressld"
        os.system(str_command)
        return
    
    with open(os.path.join(str_outputFilePath, "GenEpi_Log_" + time.strftime("%Y%m%d-%H%M", time.localtime()) + ".txt"), "w") as file_outputFile:
        ### create log
        file_outputFile.writelines("start analysis at: " + time.strftime("%Y%m%d-%H:%M:%S", time.localtime()) + "\n" + "\n")
    
        ### log arguments
        file_outputFile.writelines("Arguments in effect:" + "\n")
        file_outputFile.writelines("\t" + "-g (input genotype filename): " + str_inputFileName_genotype + "\n")
        file_outputFile.writelines("\t" + "-p (input phenotype filename): " + str_inputFileName_phenotype + "\n")
        file_outputFile.writelines("\t" + "-s (self-defined genome regions): " + str_inputFileName_regions + "\n")
        file_outputFile.writelines("\t" + "-o (output filepath): " + str_outputFilePath + "\n" + "\n")
        
        file_outputFile.writelines("\t" + "-m (model type): " + "Classification" if args.m=="c" else "Regression"  + "\n")
        file_outputFile.writelines("\t" + "-k (k-fold cross validation): " + str(args.k) + "\n")
        file_outputFile.writelines("\t" + "-t (number of threads): " + str(int_thread) + "\n" + "\n")
        
        file_outputFile.writelines("\t" + "--updatedb (enable function of update UCSC database): " + str(args.updatedb) + "\n")
        file_outputFile.writelines("\t" + "-b (human genome build): " + args.b + "\n" + "\n")
        
        file_outputFile.writelines("\t" + "--compressld (enable function of LD data compression): " + str(args.compressld) + "\n")
        file_outputFile.writelines("\t" + "-d (D prime threshold): " + str(args.d) + "\n")
        file_outputFile.writelines("\t" + "-r (R square threshold): " + str(args.r) + "\n" + "\n")
        
        ### check input format
        int_num_genotype, int_num_phenotype = InputChecking(str_inputFileName_genotype, str_inputFileName_phenotype)
        file_outputFile.writelines("Number of variants: " + str(int_num_genotype) + "\n")
        file_outputFile.writelines("Number of samples: " + str(int_num_phenotype) + "\n")
        print("Number of variants: " + str(int_num_genotype))
        print("Number of samples: " + str(int_num_phenotype))
        
        ### step1_downloadUCSCDB
        if args.updatedb:
            DownloadUCSCDB(str_hgbuild=args.b)
    
        ### step2_estimateLD
        if args.compressld:
            EstimateLDBlock(str_inputFileName_genotype, str_outputFilePath=str_outputFilePath, float_threshold_DPrime=float(args.d), float_threshold_RSquare=float(args.r))
            str_inputFileName_genotype = os.path.join(str_outputFilePath, os.path.basename(str_inputFileName_genotype.replace(".gen", "_LDReduced.gen")))
        
        ### step3_splitByGene
        if str_inputFileName_regions == "None":
            SplitByGene(str_inputFileName_genotype, str_outputFilePath=os.path.join(str_outputFilePath, "snpSubsets"))
        else:
            SplitByGene(str_inputFileName_genotype, str_inputFileName_UCSCDB=str_inputFileName_regions, str_outputFilePath=os.path.join(str_outputFilePath, "snpSubsets"))
        
        if args.m=="c":
            ### step4_singleGeneEpistasis_Logistic (for case/control trial)
            BatchSingleGeneEpistasisLogistic(os.path.join(str_outputFilePath, "snpSubsets"), str_inputFileName_phenotype, int_kOfKFold=int(args.k), int_nJobs=int(int_thread))
            ### step5_crossGeneEpistasis_Logistic (for case/control trial)
            float_score_train, float_score_test = CrossGeneEpistasisLogistic(os.path.join(str_outputFilePath, "singleGeneResult"), str_inputFileName_phenotype, int_kOfKFold=int(args.k), int_nJobs=int(int_thread))
            file_outputFile.writelines("Overall genetic feature performance (F1 score)" + "\n")
            file_outputFile.writelines("Training: " + str(float_score_train) + "\n")
            file_outputFile.writelines("Testing (" + str(args.k) + "-fold CV): " + str(float_score_test) + "\n" + "\n")
            ### step6_ensembleWithCovariates (for case/control trial)
            float_score_train, float_score_test = EnsembleWithCovariatesClassifier(os.path.join(str_outputFilePath, "crossGeneResult", "Feature.csv"), str_inputFileName_phenotype, int_kOfKFold=int(args.k), int_nJobs=int(int_thread))
            file_outputFile.writelines("Ensemble with co-variate performance (F1 score)" + "\n")
            file_outputFile.writelines("Training: " + str(float_score_train) + "\n")
            file_outputFile.writelines("Testing (" + str(args.k) + "-fold CV): " + str(float_score_test) + "\n" + "\n")
        else:
            ### step4_singleGeneEpistasis_Lasso (for quantitative trial)
            BatchSingleGeneEpistasisLasso(os.path.join(str_outputFilePath, "snpSubsets"), str_inputFileName_phenotype, int_kOfKFold=int(args.k), int_nJobs=int(int_thread))
            ### step5_crossGeneEpistasis_Lasso (for quantitative trial)
            float_score_train, float_score_test = CrossGeneEpistasisLasso(os.path.join(str_outputFilePath, "singleGeneResult"), str_inputFileName_phenotype, int_kOfKFold=int(args.k), int_nJobs=int(int_thread))
            file_outputFile.writelines("Overall genetic feature performance (Average of the Pearson and Spearman correlation)" + "\n")
            file_outputFile.writelines("Training: " + str(float_score_train) + "\n")
            file_outputFile.writelines("Testing (" + str(args.k) + "-fold CV): " + str(float_score_test) + "\n" + "\n")
            ### step6_ensembleWithCovariates (for quantitative trial)
            float_score_train, float_score_test = EnsembleWithCovariatesRegressor(os.path.join(str_outputFilePath, "crossGeneResult", "Feature.csv"), str_inputFileName_phenotype, int_kOfKFold=int(args.k), int_nJobs=int(int_thread))
            file_outputFile.writelines("Ensemble with co-variate performance (Average of the Pearson and Spearman correlation)" + "\n")
            file_outputFile.writelines("Training: " + str(float_score_train) + "\n")
            file_outputFile.writelines("Testing (" + str(args.k) + "-fold CV): " + str(float_score_test) + "\n" + "\n")
        
        file_outputFile.writelines("end analysis at: " + time.strftime("%Y%m%d-%H:%M:%S", time.localtime()) + "\n")

if __name__ == "__main__":
    main()


# -*- coding: utf-8 -*-
"""
Created on Apr 2019

@author: Chester (Yu-Chuan Chang)
"""

""""""""""""""""""""""""""""""
# import libraries
""""""""""""""""""""""""""""""
import argparse
import os
import sys

""""""""""""""""""""""""""""""
# define functions 
""""""""""""""""""""""""""""""
def ArgumentsParser():
    ### define arguments
    str_description = ''
    'This script is a preprossing tool of GenEpi for converting vcf file to oxford format'
    parser = argparse.ArgumentParser(prog='vcf2oxford', description=str_description)
    
    ### define arguments for I/O
    parser.add_argument("-v", required=True, help="filename of the input vcf file")
    parser.add_argument("-o", required=False, help="output file path")
    
    return parser

""""""""""""""""""""""""""""""
# main function
""""""""""""""""""""""""""""""
def main(args=None):
    ### obtain arguments from argument parser
    args = ArgumentsParser().parse_args(args)

    ### get arguments for I/O
    str_inputFileName_vcf = args.v
    str_outputFilePath = ""
    if args.o is not None:
        str_outputFilePath = args.o
    else:
        str_outputFilePath = os.path.dirname(str_inputFileName_vcf)

    ### count lines of input files
    int_num_vcf = sum(1 for line in open(str_inputFileName_vcf))

    ### scan .vcf file
    list_sample_id = []
    list_gen_all = []
    int_count_vcf = 0
    with open(str_inputFileName_vcf, "r") as file_inputFile:
        for line in file_inputFile:
            int_count_vcf += 1
            ### skip headers
            if line[0] == "#":
                list_sample_id = line.strip().split("\t")
            else:
                list_line = line.strip().split("\t")
                ### skip low quality variants
                if list_line[6] != "LowQual":
                    list_gen = []
                    ### chromosome
                    list_gen.append(list_line[0].replace("chr", ""))
                    ### rsID: chr|position|alt|ref
                    if list_line[2] == ".":
                        list_line[2] = list_line[0].replace("chr", "") + "|" + list_line[1] + "|" + list_line[4] + "|" + list_line[3]
                    list_gen.append(list_line[2])
                    ### position
                    list_gen.append(list_line[1])
                    ### alternative allele
                    list_gen.append(list_line[4])
                    ### reference allele
                    list_gen.append(list_line[3])
                    
                    ### encode
                    for item in list_line[9:]:
                        list_item = item.split(":")
                        if list_item[0] == ".":
                            list_gen = list_gen + ["0", "0", "1"]
                        elif list_item[0] == "0/1":
                            list_gen = list_gen + ["0", "1", "0"]
                        elif list_item[0] == "1/1":
                            list_gen = list_gen + ["1", "0", "0"]
                        else:
                            list_gen = list_gen + ["0", "0", "0"]
                    list_gen_all.append(list_gen)
            str_print = "Preprocessing vcf2oxford: " + "{0:.2f}".format(float(int_count_vcf) / int_num_vcf * 100) + "%" + "\t\t"
            sys.stdout.write('%s\r' % str_print)
            sys.stdout.flush()

    ### ouput sample file
    list_sample_id = list_sample_id[9:]
    with open(os.path.join(str_outputFilePath, str_inputFileName_vcf.replace(".vcf", ".sample")), "w") as file_outputFile:
        file_outputFile.writelines("ID_1 ID_2 missing sex phenotype" + "\n")
        file_outputFile.writelines("0 0 0 D B" + "\n")
        for item in list_sample_id:    
            file_outputFile.writelines(item + " " + item + " " + "NA 0 NA" + "\n")

    ### output gen file        
    with open(os.path.join(str_outputFilePath, str_inputFileName_vcf.replace(".vcf", ".gen")), "w") as file_outputFile:
        for item in list_gen_all:
            file_outputFile.writelines(" ".join(item) + "\n")

if __name__ == "__main__":
    main()